from syntheon.converter.converter import SynthConverter
import mido
from pathlib import Path
from syntheon.converter.dexed.dexed_constants import voice_struct, VOICE_PARAMETER_RANGES, header_struct,\
    header_bytes, voice_bytes, N_VOICES, N_OSC, KEYS


def take(take_from, n):
    for _ in range(n):
        yield next(take_from)


def checksum(data):
    return (128-sum(data)&127)%128


class DexedConverter(SynthConverter):
    def __init__(self):
        SynthConverter.__init__(self)
        self.keys = KEYS

    def serializeToDict(self, fname):
        path = Path(fname).expanduser()
        try:
            preset = mido.read_syx_file(path.as_posix())[0]
        except IndexError as e:
            return None
        except ValueError as e:
            return None
        if len(preset.data) == 0:
            return None

        def get_voice(data):
            unpacked = voice_struct.unpack(data)
            # TODO: need to take actions after verify, skip for now
            # self.verify(unpacked, VOICE_PARAMETER_RANGES)
            return unpacked

        get_header = header_struct.unpack
        sysex_iter = iter(preset.data)
        lst = []
        try:
            header = get_header(bytes(take(sysex_iter, len(header_bytes))))
            for idx in range(N_VOICES):
                x = get_voice(bytes(take(sysex_iter, len(voice_bytes)))) 
                lst.append(x)
            
            self.dict = lst
            return lst
        except RuntimeError:
            return None
    
    def parseToPluginFile(self, fname):
        def encode_head():
            header = [  '0x43',
                        '0x00',
                        '0x09',
                        '0x20',
                        '0x00',]

            return [int(i, 0) for i in header]

        def encode_osc(params, n):
            oscillator_params = []

            oscillator_params += [params[f'{n}_R1']]
            oscillator_params += [params[f'{n}_R2']]
            oscillator_params += [params[f'{n}_R3']]
            oscillator_params += [params[f'{n}_R4']]
            oscillator_params += [params[f'{n}_L1']]
            oscillator_params += [params[f'{n}_L2']]
            oscillator_params += [params[f'{n}_L3']]
            oscillator_params += [params[f'{n}_L4']]
            oscillator_params += [params[f'{n}_BP']]
            oscillator_params += [params[f'{n}_LD']]
            oscillator_params += [params[f'{n}_RD']]

            RC = params[f'{n}_RC'] << 2
            LC = params[f'{n}_LC']
            oscillator_params += [RC | LC]

            DET = params[f'{n}_DET'] << 3
            RS = params[f'{n}_RS']
            oscillator_params += [DET | RS]

            KVS = params[f'{n}_KVS'] << 2
            AMS = params[f'{n}_AMS'] 
            oscillator_params += [KVS|AMS]
            oscillator_params += [params[f'{n}_OL']]

            FC = params[f'{n}_FC'] << 1
            M = params[f'{n}_M']
            oscillator_params += [FC|M]
            oscillator_params += [params[f'{n}_FF']]

            return oscillator_params

        def encode_global(params):
            global_params = []
            global_params += [params['PR1']]
            global_params += [params['PR2']]
            global_params += [params['PR3']]
            global_params += [params['PR4']]
            global_params += [params['PL1']]
            global_params += [params['PL2']]
            global_params += [params['PL3']]
            global_params += [params['PL4']]

            global_params += [params['ALG']]

            OKS = params['OKS'] << 3
            FB = params['FB']

            global_params += [OKS|FB]
            global_params += [params['LFS']]
            global_params += [params['LFD']]
            global_params += [params['LPMD']]
            global_params += [params['LAMD']]

            LPMS = params['LPMS'] << 4
            LFW = params['LFW'] << 1
            LKS = params['LKS']
            global_params += [LPMS | LFW | LKS]
            global_params += [params['TRNSP']]
            global_params += [params[f'NAME CHAR {i + 1}'] for i in range(10)]

            return global_params

        try:
            head = encode_head()

            data = []
            assert len(self.dict) == N_VOICES

            # voices
            last_params = None
            for params in self.dict:
                if len(params.keys()) == 0:
                    params = last_params
                else:
                    last_params = params
                for osc in range(N_OSC):
                    data += encode_osc(params, osc)

                data += encode_global(params)


            this_checksum = checksum(data)
            output = [*head, *data, this_checksum]
            
            message = mido.Message('sysex', data=output)
            mido.write_syx_file(fname, [message])
            return 0
        
        except Exception as e:
            print(str(e))
            return -1
    
    def verify(self, actual, ranges):
        super().verify()
        assert set(actual.keys())==set(ranges.keys()), 'Params dont match'
        for key in actual:
            if not actual[key] in ranges[key]:
                print("returning false", key, actual[key])
                return False
        return True


if __name__ == "__main__":
    dx_converter = DexedConverter()
    dx_converter.serializeToDict("Dexed_01.syx")
    dx_converter.printMessage()
    dx_converter.parseToPluginFile("testing.syx")






from syntheon.inferencer.inferencer import Inferencer, InferenceInput, InferenceOutput
from syntheon.inferencer.dexed.models.preprocessor import ProcessData, F0LoudnessRMSPreprocessor
from syntheon.inferencer.dexed.models.ddx7.models import DDSP_Decoder, TCNFMDecoder
from syntheon.inferencer.dexed.models.ddx7.synth import FMSynth
from syntheon.inferencer.dexed.models.amp_utils import *
from syntheon.converter.dexed.dexed_converter import DexedConverter
from syntheon.utils.pitch_extractor import extract_pitch
import yaml
import torch
import librosa
import soundfile as sf
import pickle
import os
import numpy as np
import matplotlib.pyplot as plt


class DexedInferenceOutput(InferenceOutput):
    def __init__(self):
        InferenceOutput.__init__(self)
        self.synth_audio = None           # TODO: can put default values here
        self.ol = None


class DexedInferenceInput(InferenceInput):
    def __init__(self):
        self.x = None


class DexedInferencer(Inferencer):
    def convert(self, audio_fname, model_pt_fname=None, enable_eval=False):
        # TODO: convert should be more like framework. preprocess -> load_model -> inference -> post_process
        if model_pt_fname is None:
            model_pt_fname = "syntheon/inferencer/dexed/checkpoints/state_best.pth"
        with open("syntheon/inferencer/dexed/models/conf/data_config.yaml", 'r') as f:
            data_config = yaml.safe_load(f)
        
        preprocessor = ProcessData(
            silence_thresh_dB=data_config["data_processor"]["silence_thresh_dB"], 
            sr=data_config["data_processor"]["sr"], 
            device=data_config["data_processor"]["device"], 
            seq_len=data_config["data_processor"]["seq_len"],
            crepe_params=data_config["data_processor"]["crepe_params"], 
            loudness_params=data_config["data_processor"]["loudness_params"],
            rms_params=data_config["data_processor"]["rms_params"], 
            hop_size=data_config["data_processor"]["hop_size"], 
            max_len=data_config["data_processor"]["max_len"], 
            center=data_config["data_processor"]["center"]
        )

        audio, _ = librosa.load(audio_fname, sr=data_config["data_processor"]["sr"])

        f0 = extract_pitch(audio, data_config["data_processor"]["sr"], block_size=64)
        f0 = f0.astype(np.float32)
        loudness = preprocessor.calc_loudness(audio)
        rms = preprocessor.calc_rms(audio)

        scaler = F0LoudnessRMSPreprocessor()
        x = {
            "audio": torch.tensor(audio).unsqueeze(0).unsqueeze(-1),
            "f0": torch.tensor(f0).unsqueeze(0).unsqueeze(-1),
            "loudness": torch.tensor(loudness).unsqueeze(0).unsqueeze(-1),
            "rms": torch.tensor(rms).unsqueeze(0).unsqueeze(-1)
        }
        scaler.run(x)

        inference_input = DexedInferenceInput()
        inference_input.x = x

        model = self.load_model(model_pt_fname, self.device)
        inference_output = self.inference(model, inference_input, self.device, enable_eval=enable_eval)
        synth_params_dict = self.convert_to_preset(inference_output)
        return synth_params_dict, inference_output.eval_dict

    def load_model(self, model_pt_fname, device="cuda"):
        with open("syntheon/inferencer/dexed/models/conf/recipes/model/tcnres_f0ld_fmstr_noreverb.yaml", 'r') as f:
            config = yaml.safe_load(f)

        # prepare model
        decoder = TCNFMDecoder(n_blocks=config["decoder"]["n_blocks"], 
                                hidden_channels=config["decoder"]["hidden_channels"], 
                                out_channels=config["decoder"]["out_channels"],
                                kernel_size=config["decoder"]["kernel_size"],
                                dilation_base=config["decoder"]["dilation_base"],
                                apply_padding=config["decoder"]["apply_padding"],
                                deploy_residual=config["decoder"]["deploy_residual"],
                                input_keys=config["decoder"]["input_keys"])

        synth = FMSynth(sample_rate=config["synth"]["sample_rate"],
                        block_size=config["synth"]["block_size"],
                        fr=config["synth"]["fr"],
                        max_ol=config["synth"]["max_ol"],
                        synth_module=config["synth"]["synth_module"],
                        is_reverb=False)

        model = DDSP_Decoder(decoder, synth)
        if device == "cuda":
            model.load_state_dict(torch.load(model_pt_fname))
            model.cuda()
        else:
            model.load_state_dict(torch.load(model_pt_fname, map_location=torch.device('cpu')))
        model.eval()        
        return model
    
    def inference(self, model, inference_input, device="cuda", enable_eval=False):
        if device == "cuda":
            inference_input.audio = inference_input.x["audio"].cuda()
            inference_input.f0 = inference_input.x["f0"].cuda()
            inference_input.loudness = inference_input.x["loudness"].cuda()
            inference_input.rms = inference_input.x["rm"].cuda()
        
        # forward pass
        synth_out = model(inference_input.x)

        inference_output = DexedInferenceOutput()
        inference_output.synth_audio = synth_out["synth_audio"]
        inference_output.ol = synth_out["ol"]

        return inference_output
    
    def convert_to_preset(self, inference_output):

        dx_converter = DexedConverter()
        params_dict = dx_converter.serializeToDict("syntheon/inferencer/dexed/Dexed_01.syx")

        lst = []
        for idx in range(6):
            ol = inference_output.ol[0, :, idx]
            ol = ol.cpu().detach().numpy()
            ol = ol.reshape(-1, 5).mean(axis=1)

            # TODO: these are all hacky code...
            if (idx == 0 or idx == 2):
                ol = ol / 0.32
            
            lst.append(np.mean(ol))

        lst = [amplitude_to_dexed_ol(k) for k in lst]

        params_dict[0]["5_OL"] = lst[0]
        params_dict[0]["4_OL"] = lst[1]
        params_dict[0]["3_OL"] = lst[2]
        params_dict[0]["2_OL"] = lst[3]
        params_dict[0]["1_OL"] = lst[4]
        params_dict[0]["0_OL"] = lst[5]
        params_dict[0]["NAME CHAR 1"] = 83
        params_dict[0]["NAME CHAR 2"] = 89
        params_dict[0]["NAME CHAR 3"] = 78
        params_dict[0]["NAME CHAR 4"] = 84
        params_dict[0]["NAME CHAR 5"] = 72
        params_dict[0]["NAME CHAR 6"] = 69
        params_dict[0]["NAME CHAR 7"] = 79
        params_dict[0]["NAME CHAR 8"] = 78
        params_dict[0]["NAME CHAR 9"] = 32
        params_dict[0]["NAME CHAR 10"] = 32

        return params_dict


if __name__ == "__main__":
    # TODO: move to test folder
    dexed_inferencer = DexedInferencer(device="cpu")
    params = dexed_inferencer.convert("test/test_audio/dexed_test_audio_1.wav")

    from syntheon.converter.dexed.dexed_converter import DexedConverter
    dexed_converter = DexedConverter()
    dexed_converter.dict = params
    dexed_converter.parseToPluginFile("dexed_output.syx")


"""
We find it hard to map Dexed's 0-99 output level to actual amplitude.
So we conducted an empirical experiment, and manually fit the values using np.polyfit
RMS is xx. Details to be released.
"""
import numpy as np

def dexed_ol_to_amplitude(x):
    return 4e-4 * np.exp(0.086 * x)

def amplitude_to_dexed_ol(x):
    return int((np.log(x) - np.log(4e-4)) / 0.086)


"""
Diff-WTS model. Main adapted from https://github.com/acids-ircam/ddsp_pytorch.
"""
from syntheon.inferencer.vital.models.core import harmonic_synth
from syntheon.inferencer.vital.models.wavetable_synth import WavetableSynthV2
import torch
import torch.nn as nn
from syntheon.inferencer.vital.models.core import mlp, gru, scale_function, remove_above_nyquist, upsample
from syntheon.inferencer.vital.models.core import amp_to_impulse_response, fft_convolve
from syntheon.inferencer.vital.models.adsr_envelope import *
import numpy as np
from torchvision.transforms import Resize
import matplotlib.pyplot as plt
from time import time

class PrintLayer(nn.Module):
    def __init__(self, name):
        super(PrintLayer, self).__init__()
        self.name = name
    
    def forward(self, x):
        # Do your print / debug stuff here
        print(self.name, x[0].squeeze().item())
        x += 1e-2
        return x


class Reverb(nn.Module):
    def __init__(self, length, sampling_rate, initial_wet=0, initial_decay=5):
        super().__init__()
        self.length = length
        self.sampling_rate = sampling_rate

        self.noise = nn.Parameter((torch.rand(length) * 2 - 1).unsqueeze(-1))
        self.decay = nn.Parameter(torch.tensor(float(initial_decay)))
        self.wet = nn.Parameter(torch.tensor(float(initial_wet)))

        t = torch.arange(self.length) / self.sampling_rate
        t = t.reshape(1, -1, 1)
        self.register_buffer("t", t)

    def build_impulse(self):
        t = torch.exp(-nn.functional.softplus(-self.decay) * self.t * 500)
        noise = self.noise * t
        impulse = noise * torch.sigmoid(self.wet)
        impulse[:, 0] = 1
        return impulse

    def forward(self, x):
        lenx = x.shape[1]
        impulse = self.build_impulse()
        impulse = nn.functional.pad(impulse, (0, 0, 0, lenx - self.length))

        x = fft_convolve(x.squeeze(-1), impulse.squeeze(-1)).unsqueeze(-1)

        return x


def infer_wavetables(y, pitch):
    """
    TODO: do for single sample first, need to batch it later
    TODO: VERY BUGGY CODE. didn't care for edge cases (like silence). need to add the checks in

    y: (64000,)
    pitch: (400,), 1 second 100 frames
    """
    period = 1 / pitch * 16000
    
    # find the first continuous pitch
    # TODO: find the most continuous pitch across the sample, for best results
    continuous_threshold = 10   # at least 10 steps = 0.1 sec, will be a problem for plucks, but for now do it like this
    continuous_pitch = -1
    continuous_pitch_idx = -1
    cur_pitch = pitch[0]
    step = 0
    
    t1 = time()
    for idx in range(1, len(pitch)):
        if abs(pitch[idx] - cur_pitch) < 1e-2: # equal freq tolerance 1e-2
            step += 1
            if step > continuous_threshold:
                continuous_pitch = cur_pitch
                continuous_pitch_idx = idx - step
                break
        else:
            cur_pitch = pitch[idx]
            step = 0
    
    if continuous_pitch == -1:  # fallback
        continuous_pitch = pitch[0]
        continuous_pitch_idx = 0
    
    period = int(1 / continuous_pitch * 16000)
    pitch_offset_idx = continuous_pitch_idx * 160  # 160 = sr / frame_size (100)
        
    # find local minimum within a window of 2 periods
    cur = y[pitch_offset_idx : pitch_offset_idx + 1600]
    min_idx = torch.argmin(cur).item()
    
    # here we take first wavelet, but also can take the average of a few wavelets
    # TODO: prone to silence right now. need to fix. now HACK search for local minima across 1600 samples to solve
    wavelet = y[min_idx : min_idx + period]
    
    # upsample + normalize magnitude
    wavelet_tensor = wavelet.clone().detach().unsqueeze(-1).unsqueeze(0)
    if torch.isinf(wavelet_tensor).any() or torch.isnan(wavelet_tensor).any():
        print('wavelet tensor has inf or nan', torch.isinf(wavelet_tensor).any(), torch.isnan(wavelet_tensor).any())
    wavelet_upsample = upsample(wavelet_tensor, factor=0, preferred_size=512, mode="linear").squeeze()
    if torch.isinf(wavelet_upsample).any() or torch.isnan(wavelet_upsample).any():
        print('wavelet upsample has inf or nan', torch.isinf(wavelet_upsample).any(), torch.isnan(wavelet_upsample).any())
    if wavelet_upsample.max() - wavelet_upsample.min() < 1e-4:
        # don't min-max norm in this case
        pass
    else:
        wavelet_upsample = (wavelet_upsample - wavelet_upsample.min()) / \
                            (wavelet_upsample.max() - wavelet_upsample.min())
        wavelet_upsample = wavelet_upsample * 2 - 1
    if torch.isinf(wavelet_upsample).any() or torch.isnan(wavelet_upsample).any():
        print('wavelet upsample 2 has inf or nan', torch.isinf(wavelet_upsample).any(), torch.isnan(wavelet_upsample).any(),
                wavelet_upsample.max(), wavelet_upsample.min())

    return wavelet_upsample


class WTSv2(nn.Module):
    def __init__(self, hidden_size, n_harmonic, n_bands, sampling_rate,
                 block_size, mode="wavetable", duration_secs=3, num_wavetables=3,
                 wavetable_smoothing=False, min_smoothing_sigma=0.5, max_smoothing_sigma=50,
                 preload_wt=False, is_round_secs=False, enable_amplitude=True, device='cuda'
                ):
        super().__init__()
        self.register_buffer("sampling_rate", torch.tensor(sampling_rate))
        self.register_buffer("block_size", torch.tensor(block_size))

        # feature extractors
        self.encoder = mlp(30, hidden_size, 3)
        self.layer_norm = nn.LayerNorm(30)
        self.gru_mfcc = nn.GRU(30, 512, batch_first=True)
        self.mlp_mfcc = nn.Linear(512, 16)

        self.in_mlps = nn.ModuleList([mlp(1, hidden_size, 3),
                                      mlp(1, hidden_size, 3),
                                      mlp(16, hidden_size, 3)])
        self.gru = gru(3, hidden_size)
        self.out_mlp = mlp(hidden_size * 4, hidden_size, 3)

        self.loudness_mlp = nn.Sequential(
            nn.Linear(1, 1),
            nn.Sigmoid()
        )
        self.proj_matrices = nn.ModuleList([
            nn.Linear(hidden_size, n_harmonic + 1),
            nn.Linear(hidden_size, n_bands),
        ])

        # for wavetable learning
        self.wt1_conv1d = nn.Sequential(
            nn.Conv1d(1, num_wavetables, 16, stride=16),     # 3 here is num_wavetable
            nn.Tanh(),
            nn.Conv1d(num_wavetables, num_wavetables, 8, stride=8),
            nn.Tanh(),
            nn.Linear(500, 512),                # 512 is wavetable length
            nn.Tanh()
        )
        self.attention_wt1 = nn.Linear(512, 1)
        self.smoothing_linear = nn.Linear(512, 1)
        self.smoothing_sigmoid = nn.Sigmoid()

        # for adsr learning
        self.shaper = ADSREnvelopeShaper(is_round_secs)
        self.adsr_conv1d = nn.Conv1d(1, 1, block_size, stride=block_size)

        self.attack_gru = nn.GRU(1, 8, batch_first=True, bidirectional=True)
        self.decay_gru = nn.GRU(1, 8, batch_first=True, bidirectional=True)
        self.sustain_gru = nn.GRU(1, 8, batch_first=True, bidirectional=True)

        self.attack_sec_head = nn.Sequential(
            nn.Linear(16, 1),
            nn.Sigmoid()
        )
        self.decay_sec_head = nn.Sequential(
            nn.Linear(16, 1),
            nn.Sigmoid()
        )
        self.sustain_level_head = nn.Sequential(
            nn.Linear(16, 1),
            nn.Sigmoid()
        )

        # for adsr result storage
        self.attack_sec = nn.Parameter(torch.ones(1,))
        self.decay_sec = nn.Parameter(torch.ones(1,))
        self.sustain_level = nn.Parameter(torch.ones(1,))       

        self.max_attack_secs = 2.0
        self.max_decay_secs = 2.0 

        # for synthesis
        self.reverb = Reverb(sampling_rate, sampling_rate)
        self.wts = WavetableSynthV2(sr=sampling_rate, 
                                    duration_secs=duration_secs,
                                    block_size=block_size,
                                    enable_amplitude=enable_amplitude)
        self.wavetable_smoothing = wavetable_smoothing
        self.min_smoothing_sigma = min_smoothing_sigma
        self.max_smoothing_sigma = max_smoothing_sigma

        self.preload_wt = preload_wt

        self.mode = mode
        self.duration_secs = duration_secs
        self.device = device

    def forward(self, y, mfcc, pitch, loudness, times, onset_frames):
        batch_size = y.shape[0]

        # encode mfcc first
        # use layer norm instead of trainable norm, not much difference found
        mfcc = self.layer_norm(torch.transpose(mfcc, 1, 2))
        mfcc = self.gru_mfcc(mfcc)[0]
        mfcc = self.mlp_mfcc(mfcc)

        # use image resize to align dimensions, ddsp also do this...
        mfcc = Resize(size=(self.duration_secs * 100, 16))(mfcc)

        hidden = torch.cat([
            self.in_mlps[0](pitch),
            self.in_mlps[1](loudness),
            self.in_mlps[2](mfcc)
        ], -1)
        hidden = torch.cat([self.gru(hidden)[0], hidden], -1)
        hidden = self.out_mlp(hidden)

        # harmonic part
        total_amp = self.loudness_mlp(loudness)
        pitch_prev = pitch

        # TODO: upsample is very slow
        pitch = upsample(pitch, self.block_size)
        total_amp = upsample(total_amp, self.block_size)    # use this instead for wavetable

        # diff-wave-synth synthesizer
        if self.preload_wt:
            # TODO: very slow implementation...
            wavetables = []
            for idx in range(batch_size):
                wt = infer_wavetables(y[idx].squeeze(), pitch_prev[idx].squeeze())
                wavetables.append(wt)
            wavetables = torch.stack(wavetables, dim=0).unsqueeze(1)
            if torch.isinf(wavetables).any() or torch.isnan(wavetables).any():
                print('wavetables has inf or nan', torch.isinf(wavetables).any(), torch.isnan(wavetables).any())
        else:
            wavetables = self.wt1_conv1d(y.unsqueeze(1))

        if self.wavetable_smoothing:
            smoothing_coeff = self.smoothing_linear(wavetables)
            smoothing_coeff = smoothing_coeff.squeeze(1)        # HACK: here should assume only 1 wavetable
            smoothing_coeff = self.smoothing_sigmoid(smoothing_coeff)
            wavetables_old = wavetables
            wavetables = self.smoothing(wavetables, smoothing_coeff)
        else:
            wavetables_old = None
            smoothing_coeff = None

        attention_output = self.attention_wt1(wavetables).squeeze(-1)
        attention_output = nn.Softmax(dim=-1)(attention_output)

        harmonic, attention_output = self.wts(pitch, total_amp, wavetables, attention_output)

        # noise part
        param = scale_function(self.proj_matrices[1](hidden) - 5)

        impulse = amp_to_impulse_response(param, self.block_size)
        noise = torch.rand(
            impulse.shape[0],
            impulse.shape[1],
            self.block_size,
        ).to(impulse) * 2 - 1

        noise = fft_convolve(noise, impulse).contiguous()
        noise = noise.reshape(noise.shape[0], -1, 1)

        signal = harmonic + noise

        # adsr shaping
        output_attack, hn_attack = self.attack_gru(loudness)
        hn_attack = torch.cat([hn_attack[0], hn_attack[1]], dim=-1)
        output_decay, hn_decay = self.decay_gru(loudness)
        hn_decay = torch.cat([hn_decay[0], hn_decay[1]], dim=-1)
        output_sustain, hn_sustain = self.sustain_gru(loudness)
        hn_sustain = torch.cat([hn_sustain[0], hn_sustain[1]], dim=-1)

        # print(hn_decay[:10])
        attack_level = self.attack_sec_head(hn_attack).squeeze()            # 0-1
        decay_level = self.decay_sec_head(hn_decay).squeeze()               # 0-1
        sustain_level = self.sustain_level_head(hn_sustain).squeeze()

        attack_secs = attack_level * self.max_attack_secs
        decay_secs = decay_level * self.max_decay_secs
        # print(attack_secs, decay_secs, sustain_level)
        # attack_secs = torch.tensor([0.0, 0.0]).cuda()
        # decay_secs = torch.tensor([0.1, 0.1]).cuda()
        # sustain_level = torch.tensor([0.001, 0.001]).cuda()

        amp_onsets = np.append(times[onset_frames], np.array([times[-1]]))  # TODO: now 1 onset is enough, because all training samples pitch are the same

        adsr = get_amp_shaper_v2(self.shaper, amp_onsets, 
                                attack_secs=attack_secs,
                                decay_secs=decay_secs,
                                sustain_level=sustain_level)
        if adsr.shape[1] < pitch_prev.shape[1]:
            # adsr = torch.nn.functional.pad(adsr, (0, pitch_prev.shape[1] - adsr.shape[1]), "constant", adsr[-1].item())
            adsr = torch.cat([adsr, adsr[:, -1].unsqueeze(-1)], dim=-1)
        else:
            adsr = adsr[:pitch_prev.shape[1]]
        
        self.adsr = adsr

        adsr_prev = adsr
        adsr = adsr.unsqueeze(-1)
        adsr = upsample(adsr, self.block_size).squeeze(-1)   

        adsr = adsr[:, :signal.shape[1]]

        final_signal = signal.squeeze() * adsr

        # final_signal = signal

        # reverb part
        # signal = self.reverb(signal)

        return signal, (attack_secs, decay_secs, sustain_level), final_signal, attention_output, wavetables, wavetables_old, smoothing_coeff
        # return signal, adsr_prev, final_signal, attention_output, wavetables, wavetables_old, smoothing_coeff
    
    def smoothing_deprecate(self, wavetables, p):
        """
        wavetables: size (b, wavetable_length)
        p: size (b,). value between 0-1
        """
        wavetables = wavetables.squeeze()   # HACK: also a hack, because by right wavetables.dim[1] is num_wavetables, here assume 1 wavetable
        bs, wavetable_length = wavetables.shape[0], wavetables.shape[1]
        smoothed_wavetables = torch.zeros((bs, wavetable_length))
        if self.device == "cuda":
            smoothed_wavetables = smoothed_wavetables.cuda()

        sigma = p * (self.max_smoothing_sigma - self.min_smoothing_sigma) + self.min_smoothing_sigma

        x_vals = torch.arange(wavetable_length)
        if self.device == "cuda":
            x_vals = x_vals.cuda()
        x_vals = torch.stack([x_vals] * bs, dim=0)

        # TODO: can we not do for loop here?
        for x_position in range(wavetable_length):
            kernel = torch.exp(-(x_vals - x_position) ** 2 / (2 * sigma ** 2))
            kernel = kernel / torch.sum(kernel, dim=1).unsqueeze(-1)
            smoothed_wavetables[:, x_position] = torch.sum(wavetables * kernel, dim=-1)
    
        return smoothed_wavetables.unsqueeze(1)
    
    def smoothing(self, wavetables, p):       
        bs, wavetable_length = wavetables.shape[0], wavetables.shape[2]
        smoothed_wavetables = torch.zeros((bs, wavetable_length))
        if self.device == "cuda":
            smoothed_wavetables = smoothed_wavetables.cuda()
        
        sigma = p * (self.max_smoothing_sigma - self.min_smoothing_sigma) + self.min_smoothing_sigma
        sigma = sigma.unsqueeze(-1)                         # size (bs, 1, 1)
        
        kernel = torch.arange(wavetable_length)
        if self.device == "cuda":
            kernel = kernel.cuda()
        kernel = kernel.unsqueeze(0) - kernel.unsqueeze(-1)  # x_position - x_vals, size (wt_len, wt_len)

        kernel = torch.exp(-kernel ** 2 / (2 * sigma ** 2))   # size (b, wt_len, wt_len)    
        kernel = kernel / torch.sum(kernel, dim=-1).unsqueeze(-1)          # dim 1 or dim -1?
            
        # wavetables = wavetables.unsqueeze(1)
        smoothed_wavetables = torch.bmm(wavetables, kernel)  # (bs, 1, wt_len) * (bs, wt_len, wt_len)
        return smoothed_wavetables


import torch
import torch.nn.functional as F
import torchcrepe
import torchaudio
import librosa
from syntheon.inferencer.dexed.models.ddx7.core import _DB_RANGE,_REF_DB
import math
import numpy as np
from time import time

_RMS_FRAME = 2048
_CREPE_WIN_LEN = 1024
_LD_N_FFT = 2048

def safe_log(x):
    return torch.log(x + 1e-7)

def calc_f0(audio, rate, hop_size,fmin,fmax,model,
            batch_size,device,center=False):
    if center is False:
      # Add padding to the end. Then execute crepe w/o padding.
      # Crepe pads so that the signal stays in the center.
      n_samples_initial = int(audio.shape[-1])
      n_frames = int(np.ceil(n_samples_initial / hop_size))
      n_samples_final = (n_frames - 1) * hop_size + _CREPE_WIN_LEN
      pad = n_samples_final - n_samples_initial
      audio = np.pad(audio, ((0, pad),), "constant")

    audio = torch.from_numpy(audio).unsqueeze(0).float().to(device)

    t1 = time()
    print("predicting...")
    crepe_tuple = torchcrepe.predict(audio,
                        rate,
                        hop_size,
                        fmin,
                        fmax,
                        model,
                        return_periodicity=True,
                        batch_size=batch_size,
                        device=device,
                        pad=center)
    print("done...", time() - t1)

    f0 = crepe_tuple[0]
    confidence = crepe_tuple[1]
    if center is True:
      f0 = f0[:,0:-1] #Discard the last sample
      confidence = confidence[:,0:-1] #Discard the last sample

    f0 = f0.squeeze(0).cpu().numpy()
    confidence = confidence.squeeze(0).cpu().numpy()
    return f0,confidence

def calc_loudness(audio, rate, n_fft=_LD_N_FFT, hop_size=64,
                  range_db=_DB_RANGE,ref_db=_REF_DB,center=False):
    np.seterr(divide='ignore')

    """Compute loudness, add to example (ref is white noise, amplitude=1)."""
    # Copied from magenta/ddsp/spectral_ops.py
    # Get magnitudes.
    if center is False:
        # Add padding to the end
        n_samples_initial = int(audio.shape[-1])
        n_frames = int(np.ceil(n_samples_initial / hop_size))
        n_samples_final = (n_frames - 1) * hop_size + n_fft
        pad = n_samples_final - n_samples_initial
        audio = np.pad(audio, ((0, pad),), "constant")
    spectra = librosa.stft(
        audio, n_fft=n_fft, hop_length=hop_size, center=center).T

    # Compute power
    amplitude = np.abs(spectra)
    amin = 1e-20  # Avoid log(0) instabilities.
    power_db = np.log10(np.maximum(amin, amplitude))
    power_db *= 20.0

    # Perceptual weighting.
    frequencies = librosa.fft_frequencies(sr=rate, n_fft=n_fft)
    a_weighting = librosa.A_weighting(frequencies)[np.newaxis, :]
    loudness = power_db + a_weighting

    # Set dynamic range.
    loudness -= ref_db
    loudness = np.maximum(loudness, -range_db)

    # Average over frequency bins. (loudness is taken from the fft dimension!)
    mean_loudness_db = np.mean(loudness, axis=-1)
    return mean_loudness_db.astype(np.float32)

'''
RMS POWER COMPUTATION.
'''

def amplitude_to_db(amplitude):
  """Converts amplitude to decibels."""
  amin = 1e-20  # Avoid log(0) instabilities.
  db = np.log10(np.maximum(amin, amplitude))
  db *= 20.0
  return db

def compute_rms_energy(audio,
                       frame_size=2048,
                       hop_size=64,
                       pad_end=True):
  """Compute root mean squared energy of audio."""
  if pad_end is True:
    # Add padding to the end
    n_samples_initial = int(audio.shape[-1])
    n_frames = int(np.ceil(n_samples_initial / hop_size))
    n_samples_final = (n_frames - 1) * hop_size + frame_size
    pad = n_samples_final - n_samples_initial
    audio = np.pad(audio, ((0, pad),), "constant")

  audio = torch.tensor(audio)
  audio_frames = audio.unfold(-1,frame_size,hop_size)
  rms_energy = torch.mean(audio_frames**2.0,dim=-1)**0.5

  return rms_energy.cpu().numpy()


def calc_power(audio,
                frame_size=_RMS_FRAME,
                hop_size=64,
                range_db=_DB_RANGE,
                ref_db=20.7,
                pad_end=True):
  """Compute power of audio in dB."""
  rms_energy = compute_rms_energy(audio, frame_size, hop_size,pad_end=pad_end)
  power_db = amplitude_to_db(rms_energy**2)
  #print(power_db)
  # Set dynamic range.
  power_db -= ref_db
  power_db = np.maximum(power_db, -range_db)
  return power_db.astype(np.float32)



import json
import base64
import struct
from syntheon.converter.converter import SynthConverter
from syntheon.converter.vital.vital_constants import N_WAVETABLES, CUSTOM_KEYS
import numpy as np
import math


class Base64Converter:
    def __init__(self):
        pass

    def encode(self, signal):
        signal_bytes = struct.pack('{}f'.format(len(signal)), *signal)
        base64_string = base64.b64encode(signal_bytes)

        return base64_string.decode('ascii')
    
    def decode(self, base64_string, output_length=2048):
        signal_bytes = base64.decodebytes(base64_string.encode('ascii'))
        arr = [k for k in struct.iter_unpack('f', signal_bytes)]    # unpack as 2 bytes integer
        arr = [k[0] for k in arr]                                   # normalize

        return np.array(arr)


class VitalConverter(SynthConverter):
    def __init__(self):
        SynthConverter.__init__(self)
        self.keys = []
        self.base64_converter = Base64Converter()
    
    def serializeToDict(self, fname):
        try:
            with open(fname) as f:
                self.dict = json.load(f)
            
            # decode custom part
            self.dict[CUSTOM_KEYS] = {}
            self.dict[CUSTOM_KEYS]["wavetables"] = []
            for idx in range(len(N_WAVETABLES)):
                wavetable_str = self.dict["settings"]["wavetables"][idx]["groups"][0]["components"][0]["keyframes"][0]["wave_data"]
                wavetable_name = self.dict["settings"]["wavetables"][idx]["name"]
                wavetable_osc_level = self.dict["settings"]["osc_{}_level".format(idx + 1)]
                wavetable = self.base64_converter.decode(wavetable_str)     # return np.array
                cur_dict = {
                    "name": wavetable_name,
                    "wavetable": wavetable,
                    "osc_level": wavetable_osc_level
                }
                self.dict[CUSTOM_KEYS]["wavetables"].append(cur_dict)
            
            # switch off unused wavetables
            if N_WAVETABLES == 1:
                self.dict["settings"]["osc_2_on"] = 0.0
                self.dict["settings"]["osc_3_on"] = 0.0
            elif N_WAVETABLES == 2:
                self.dict["settings"]["osc_3_on"] = 0.0
                    
        except Exception as e:
            print(str(e))
        
        return self.dict
    
    def parseToPluginFile(self, fname):
        """
        vital parameters value scale: https://github.com/mtytel/vital/blob/c0694a193777fc97853a598f86378bea625a6d81/src/common/synth_parameters.cpp
        value scale computation: https://github.com/mtytel/vital/blob/c0694a193777fc97853a598f86378bea625a6d81/src/plugin/value_bridge.h
        """
        # encode custom part
        wavetables = self.dict[CUSTOM_KEYS]["wavetables"]
        for idx in range(N_WAVETABLES):
            wavetable = wavetables[idx]["wavetable"]
            wavetable_name = wavetables[idx]["name"]
            wavetable_osc_level = wavetables[idx]["osc_level"]

            wavetable_str = self.base64_converter.encode(wavetable)
            self.dict["settings"]["wavetables"][idx]["groups"][0]["components"][0]["keyframes"][0]["wave_data"] = wavetable_str
            self.dict["settings"]["wavetables"][idx]["name"] = wavetable_name
            self.dict["settings"]["osc_{}_level".format(idx + 1)] = wavetable_osc_level
        
        # switch off unused wavetables
        if N_WAVETABLES == 1:
            self.dict["settings"]["osc_2_on"] = 0.0
            self.dict["settings"]["osc_3_on"] = 0.0
        elif N_WAVETABLES == 2:
            self.dict["settings"]["osc_3_on"] = 0.0
        
        # adsr filter
        adsrs = self.dict[CUSTOM_KEYS]["adsr"]
        # attack is kQuartic
        self.dict["settings"]["env_1_attack"] = math.sqrt(math.sqrt(adsrs["attack"]))
        # attack power is kLinear
        self.dict["settings"]["env_1_attack_power"] = adsrs["attack_power"]
        # decay is kQuartic
        self.dict["settings"]["env_1_decay"] = math.sqrt(math.sqrt(adsrs["decay"]))
        # decay power is kLinear
        self.dict["settings"]["env_1_decay_power"] = adsrs["decay_power"]
        # sustain is kLinear
        self.dict["settings"]["env_1_sustain"] = adsrs["sustain"]

        # self.dict["settings"]["env_1_delay"] = adsrs["delay"]
        # self.dict["settings"]["env_1_hold"] = adsrs["hold"]
        # self.dict["settings"]["env_1_release"] = adsrs["release"]
        # self.dict["settings"]["env_1_release_power"] = adsrs["release_power"]
        # y["settings"]["lfos"] = x_init["settings"]["lfos"]

        del self.dict[CUSTOM_KEYS]

        with open(fname ,"w+") as f:
            json.dump(self.dict, f)


import numpy as np
import librosa
import soundfile as sf


# helper functions to generate wavetable
def generate_wavetable(length, f):
    wavetable = np.zeros((length,), dtype=np.float32)
    for i in range(length):
        wavetable[i] = f(2 * np.pi * i / length)
    return wavetable


def sawtooth_waveform(x):
    """Sawtooth with period 2 pi."""
    return (x + np.pi) / np.pi % 2 - 1


def square_waveform(x):
    """Square waveform with period 2 pi."""
    return np.sign(np.sin(x))


def trim_audio(in_name, out_name, start_sec, end_sec, sr=44100):
    x, sr = librosa.load(in_name, sr=sr)
    x = x[start_sec * sr: end_sec * sr]
    sf.write(out_name, x, sr, 'PCM_24')


if __name__ == "__main__":
    trim_audio("test_audio/kygo_pluck.mp3", "kygo_pluck.wav", 75, 85)


class SynthConverter:
    
    def __init__(self):
        self.dict = None
        self.keys = []      # keys that need to be filled for this synth
    
    def serializeToDict(self, fname):
        """
        From plugin file to protobuf.

        Args:
        fname - input file name
        """
        return None
    
    def parseToPluginFile(self, fname):
        """
        From protobuf to plugin file.

        Args:
        fname - output file name
        """
        return None
    
    def printMessage(self):
        """
        Print synth parameters.
        """
        if self.dict:
            print(self.dict)
        
        else:
            raise ValueError("synth parameters not serialized yet")
    
    def keys(self):
        return self.keys
    
    def verify(self):
        """
        Verify if params are valid. Used in serializeToDict method.
        """
        if self.dict is None:
            raise ValueError("synth parameters not serialized yet")
        
        # value range checks can leave to derived classes
        for key in self.keys:
            if isinstance(self.dict, list):
                for elem in self.dict:
                    if key not in elem:
                        raise ValueError("specified key not in synth parameters: {}".format(key))
            elif isinstance(self.dict, dict):
                if key not in elem:
                        raise ValueError("specified key not in synth parameters: {}".format(key))


"""
Connects model output to synth preset parameter IR.
"""
class InferenceInput:
    def __init__(self):
        return NotImplementedError


class InferenceOutput:
    def __init__(self):
        # for storing evaluation results
        self.eval_dict = {
            "loss":  -1
        }


class Inferencer:
    def __init__(self, device="cuda"):
        self.device = device

    def convert(self, model_pt_fname, audio_fname):
        model = self.load_model(model_pt_fname, self.device)
        inference_output = self.inference(model, audio_fname, self.device)
        synth_params_dict = self.convert_to_preset(inference_output)
        return synth_params_dict, inference_output.eval_dict
    
    def load_model(self, model_pt_fname, device="cuda"):
        return NotImplementedError
        
    def inference(self, model, audio_fname):
        return NotImplementedError

    def convert_to_preset(self, inference_output):
        return NotImplementedError


import torch
import torch.nn as nn
import math
from syntheon.inferencer.dexed.models.ddx7.core import *
import matplotlib.pyplot as plt
import soundfile as sf
import librosa

def exp_sigmoid(x):
    return 2 * torch.sigmoid(x)**(math.log(10)) + 1e-7

def remove_above_nyquist(amplitudes, pitch, sampling_rate):
    n_harm = amplitudes.shape[-1]
    pitches = pitch * torch.arange(1, n_harm + 1).to(pitch)
    aa = (pitches < sampling_rate / 2).float() + 1e-4
    return amplitudes * aa


class FMSynth(nn.Module):
    def __init__(self,sample_rate,block_size,fr=[1,1,1,1,3,14],max_ol=2,
        scale_fn = torch.sigmoid,synth_module='fmstrings',is_reverb=True):
        super().__init__()
        self.sample_rate = sample_rate
        self.block_size = block_size
        self.reverb = Reverb(length=sample_rate, sample_rate=sample_rate)
        fr = torch.tensor(fr) # Frequency Ratio
        self.register_buffer("fr", fr) #Non learnable but sent to GPU if declared as buffers, and stored in model dictionary
        self.scale_fn = scale_fn
        self.use_cumsum_nd = False
        self.max_ol = max_ol
        self.is_reverb = is_reverb

        available_synths = {
            'fmbrass': fm_brass_synth,
            'fmflute': fm_flute_synth,
            'fmstrings': fm_string_synth,
            'fmablbrass': fm_ablbrass_synth,
            '2stack2': fm_2stack2,
            '1stack2':fm_1stack2,
            '1stack4': fm_1stack4}

        self.synth_module = available_synths[synth_module]

    def forward(self,controls):

        ol = self.max_ol*self.scale_fn(controls['ol'])
        ol_up = upsample(ol, self.block_size,'linear')
        f0_up = upsample(controls['f0_hz'], self.block_size,'linear')
        signal = self.synth_module(f0_up,
                                ol_up,
                                self.fr,
                                self.sample_rate,
                                self.max_ol,
                                self.use_cumsum_nd)
        #reverb part
        if self.is_reverb:
            signal = self.reverb(signal)

        synth_out = {
            'synth_audio': signal,
            'ol': ol,
            'f0_hz': controls['f0_hz']
            }
        return synth_out

class HNSynth(nn.Module):
    def __init__(self,sample_rate,block_size,scale_fn = exp_sigmoid):
        super().__init__()
        self.sample_rate = sample_rate
        self.block_size = block_size
        self.reverb = Reverb(length=sample_rate, sample_rate=sample_rate)
        self.use_cumsum_nd = False
        self.scale_fn = scale_fn

    # expects: harmonic_distr, amplitude, noise_bands
    def forward(self,controls):

        harmonics = self.scale_fn(controls['harmonic_distribution'])
        noise_bands = self.scale_fn(controls['noise_bands'])
        total_amp = self.scale_fn(controls['amplitude'])

        harmonics = remove_above_nyquist(
            harmonics,
            controls['f0_hz'],
            self.sample_rate,
        )
        harmonics /= harmonics.sum(-1, keepdim=True)
        harmonics *= total_amp

        harmonics_up = upsample(harmonics, self.block_size)
        f0_up = upsample(controls['f0_hz'], self.block_size,'linear')

        harmonic = harmonic_synth(f0_up, harmonics_up, self.sample_rate, self.use_cumsum_nd)
        impulse = amp_to_impulse_response(noise_bands, self.block_size)

        noise = torch.rand(
            impulse.shape[0],
            impulse.shape[1],
            self.block_size,
            ).to(impulse) * 2 - 1

        noise = fft_convolve(noise, impulse).contiguous()
        noise = noise.reshape(noise.shape[0], -1, 1)

        signal = harmonic + noise

        #reverb part
        signal = self.reverb(signal)
        synth_out = {
            'synth_audio': signal,
            'harmonic_distribution': harmonics,
            'noise_bands': noise_bands,
            'f0_hz': controls['f0_hz']
            }

        return synth_out

class Reverb(nn.Module):
    def __init__(self, length, sample_rate, initial_wet=0, initial_decay=5):
        super().__init__()
        self.length = length
        self.sample_rate = sample_rate

        self.noise = nn.Parameter((torch.rand(length) * 2 - 1).unsqueeze(-1))
        self.decay = nn.Parameter(torch.tensor(float(initial_decay)))
        self.wet = nn.Parameter(torch.tensor(float(initial_wet)))

        t = torch.arange(self.length) / self.sample_rate
        t = t.reshape(1, -1, 1)
        self.register_buffer("t", t)

    def build_impulse(self):
        t = torch.exp(-nn.functional.softplus(-self.decay) * self.t * 500)
        noise = self.noise * t
        impulse = noise * torch.sigmoid(self.wet)
        impulse[:, 0] = 1
        return impulse

    def forward(self, x):
        lenx = x.shape[1]
        impulse = self.build_impulse()
        impulse = nn.functional.pad(impulse, (0, 0, 0, lenx - self.length))

        x = fft_convolve(x.squeeze(-1), impulse.squeeze(-1)).unsqueeze(-1)

        return x


if __name__ == "__main__":
    fmsynth_string = FMSynth(is_reverb=False, sample_rate=16000, block_size=64)
    freq = 523.25
    controls = {}
    controls['f0_hz'] = torch.ones(1, 1000, 1) * freq
    controls['ol'] = torch.zeros(1, 1000, 6)

    synth_out = fmsynth_string(controls)['synth_audio']

    signal = synth_out.squeeze().cpu().detach().numpy()
    # signal_gt, sr = librosa.load("dexed_output_ol50_coarse1.wav", sr=16000)
    # signal_gt = signal_gt / np.amax(signal_gt)
    # # print(signal_gt.shape)

    # plt.plot(signal[16000:16400], label="signal")
    # plt.plot(signal_gt[16001:16401], label="signal_gt")
    # plt.legend()
    # plt.show()

    # print(np.amax(signal_gt))

    # S_dexed = np.abs(librosa.stft(signal_gt))
    # S_test = np.abs(librosa.stft(signal))

    # fig, ax = plt.subplots()
    # import librosa.display
    # img = librosa.display.specshow(librosa.amplitude_to_db(S_dexed,
    #                                                     ref=np.max),
    #                             y_axis='log', x_axis='time', ax=ax)
    # ax.set_title('Dexed')
    # fig.colorbar(img, ax=ax, format="%+2.0f dB")
    # plt.show()

    # fig, ax = plt.subplots()
    # img = librosa.display.specshow(librosa.amplitude_to_db(S_test,
    #                                                     ref=np.max),
    #                             y_axis='log', x_axis='time', ax=ax)
    # ax.set_title('Test')
    # fig.colorbar(img, ax=ax, format="%+2.0f dB")
    # plt.show()

    sf.write("dexed_test_92.wav", signal, 16000)




"""
Diff-WTS model. Main adapted from https://github.com/acids-ircam/ddsp_pytorch.
"""
from syntheon.inferencer.vital.models.core import harmonic_synth
from syntheon.inferencer.vital.models.wavetable_synth import WavetableSynth
import torch
import torch.nn as nn
from syntheon.inferencer.vital.models.core import mlp, gru, scale_function, remove_above_nyquist, upsample
from syntheon.inferencer.vital.models.core import amp_to_impulse_response, fft_convolve
from syntheon.inferencer.vital.models.adsr_envelope import *
import numpy as np
from torchvision.transforms import Resize
import matplotlib.pyplot as plt


class Reverb(nn.Module):
    def __init__(self, length, sampling_rate, initial_wet=0, initial_decay=5):
        super().__init__()
        self.length = length
        self.sampling_rate = sampling_rate

        self.noise = nn.Parameter((torch.rand(length) * 2 - 1).unsqueeze(-1))
        self.decay = nn.Parameter(torch.tensor(float(initial_decay)))
        self.wet = nn.Parameter(torch.tensor(float(initial_wet)))

        t = torch.arange(self.length) / self.sampling_rate
        t = t.reshape(1, -1, 1)
        self.register_buffer("t", t)

    def build_impulse(self):
        t = torch.exp(-nn.functional.softplus(-self.decay) * self.t * 500)
        noise = self.noise * t
        impulse = noise * torch.sigmoid(self.wet)
        impulse[:, 0] = 1
        return impulse

    def forward(self, x):
        lenx = x.shape[1]
        impulse = self.build_impulse()
        impulse = nn.functional.pad(impulse, (0, 0, 0, lenx - self.length))

        x = fft_convolve(x.squeeze(-1), impulse.squeeze(-1)).unsqueeze(-1)

        return x


class WTS(nn.Module):
    def __init__(self, hidden_size, n_harmonic, n_bands, sampling_rate,
                 block_size, n_wavetables, mode="wavetable", duration_secs=3,
                 adsr_hidden_size=1200,
                ):
        super().__init__()
        self.register_buffer("sampling_rate", torch.tensor(sampling_rate))
        self.register_buffer("block_size", torch.tensor(block_size))

        self.encoder = mlp(30, hidden_size, 3)
        self.layer_norm = nn.LayerNorm(30)
        self.gru_mfcc = nn.GRU(30, 512, batch_first=True)
        self.mlp_mfcc = nn.Linear(512, 16)

        self.in_mlps = nn.ModuleList([mlp(1, hidden_size, 3),
                                      mlp(1, hidden_size, 3),
                                      mlp(16, hidden_size, 3)])
        self.gru = gru(3, hidden_size)
        self.out_mlp = mlp(hidden_size * 4, hidden_size, 3)

        self.loudness_mlp = nn.Sequential(
            nn.Linear(1, 1),
            nn.Sigmoid()
        )

        self.proj_matrices = nn.ModuleList([
            nn.Linear(hidden_size, n_harmonic + 1),
            nn.Linear(hidden_size, n_bands),
        ])

        # for adsr learning
        self.adsr_conv1d = nn.Conv1d(1, 1, block_size, stride=block_size)
        self.attack_sec_head = nn.Sequential(
            nn.Linear(300, 1),
            nn.ReLU()
        )
        self.decay_sec_head = nn.Sequential(
            nn.Linear(300, 1),
            nn.ReLU()
        )
        self.sustain_level_head = nn.Sequential(
            nn.Linear(300, 1),
            nn.Sigmoid()
        )

        self.reverb = Reverb(sampling_rate, sampling_rate)
        self.wts = WavetableSynth(n_wavetables=n_wavetables, 
                                  sr=sampling_rate, 
                                  duration_secs=duration_secs,
                                  block_size=block_size)

        self.register_buffer("cache_gru", torch.zeros(1, 1, hidden_size))
        self.register_buffer("phase", torch.zeros(1))

        self.mode = mode
        self.duration_secs = duration_secs
        self.adsr = None

        self.shaper = ADSREnvelopeShaper()

    def forward(self, y, mfcc, pitch, loudness, times, onset_frames):
        # encode mfcc first
        # use layer norm instead of trainable norm, not much difference found
        mfcc = self.layer_norm(torch.transpose(mfcc, 1, 2))
        mfcc = self.gru_mfcc(mfcc)[0]
        mfcc = self.mlp_mfcc(mfcc)

        # use image resize to align dimensions, ddsp also do this...
        mfcc = Resize(size=(self.duration_secs * 100, 16))(mfcc)

        hidden = torch.cat([
            self.in_mlps[0](pitch),
            self.in_mlps[1](loudness),
            self.in_mlps[2](mfcc)
        ], -1)
        hidden = torch.cat([self.gru(hidden)[0], hidden], -1)
        hidden = self.out_mlp(hidden)

        # harmonic part
        param = self.proj_matrices[0](hidden)
        total_amp = param[..., :1]
        amplitudes = param[..., 1:]

        amplitudes = remove_above_nyquist(
            amplitudes,
            pitch,
            self.sampling_rate,
        )
        amplitudes /= amplitudes.sum(-1, keepdim=True)
        amplitudes *= total_amp

        total_amp_2 = self.loudness_mlp(loudness)

        amplitudes = upsample(amplitudes, self.block_size)
        pitch_prev = pitch
        pitch = upsample(pitch, self.block_size)
        total_amp = upsample(total_amp, self.block_size)    # TODO: wts can't backprop when using this total_amp, not sure why
        total_amp_2 = upsample(total_amp_2, self.block_size)    # use this instead for wavetable

        # diff-wave-synth synthesizer
        harmonic, attention_output = self.wts(pitch, total_amp_2)

        # noise part
        param = scale_function(self.proj_matrices[1](hidden) - 5)

        impulse = amp_to_impulse_response(param, self.block_size)
        noise = torch.rand(
            impulse.shape[0],
            impulse.shape[1],
            self.block_size,
        ).to(impulse) * 2 - 1

        noise = fft_convolve(noise, impulse).contiguous()
        noise = noise.reshape(noise.shape[0], -1, 1)

        signal = harmonic + noise

        # adsr shaping
        y_hidden = self.adsr_conv1d(y.unsqueeze(1))
        attack_secs = self.attack_sec_head(y_hidden.squeeze()).squeeze()
        decay_secs = self.decay_sec_head(y_hidden.squeeze()).squeeze()
        sustain_level = self.sustain_level_head(y_hidden.squeeze()).squeeze()

        amp_onsets = np.append(times[onset_frames], np.array([times[-1]]))

        adsr = get_amp_shaper(self.shaper, amp_onsets, 
                            attack_secs=attack_secs,
                            decay_secs=decay_secs,
                            sustain_level=sustain_level)
        if adsr.shape[0] < pitch_prev.shape[1]:
            adsr = torch.nn.functional.pad(adsr, (0, pitch_prev.shape[1] - adsr.shape[0]), "constant", adsr[-1].item())
        else:
            adsr = adsr[:pitch_prev.shape[1]]
        
        self.adsr = adsr

        adsr = adsr.unsqueeze(0).unsqueeze(-1)
        adsr = upsample(adsr, self.block_size).squeeze().unsqueeze(-1)

        # temporary: fix adsr for more examples
        adsr = torch.stack([adsr] * signal.shape[0], dim=0)
        final_signal = signal * adsr
        # final_signal = signal

        # reverb part
        # signal = self.reverb(signal)

        return signal, adsr, final_signal, attention_output


import torch
import operator
import functools
import h5py
from pathlib import Path
from tqdm import tqdm
import numpy as np
import librosa
from syntheon.inferencer.dexed.models.ddx7 import spectral_ops
from syntheon.inferencer.dexed.models.ddx7.core import hz_to_midi, _DB_RANGE, _F0_RANGE


class dotdict(dict):
    """dot.notation access to dictionary attributes"""
    __getattr__ = dict.get
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__


class ProcessData():
    def __init__(self, silence_thresh_dB, sr, device, seq_len,
                crepe_params, loudness_params,
                rms_params, hop_size, max_len, center,
                overlap = 0.0,
                debug = False,
                contiguous = False,
                contiguous_clip_noise = False):
        super().__init__()
        self.silence_thresh_dB = silence_thresh_dB
        self.crepe_params = crepe_params
        self.sr = sr
        self.device = torch.device(device)
        self.seq_len = seq_len
        self.loudness_params = loudness_params
        self.rms = rms_params
        self.max_len = max_len
        self.hop_size = hop_size
        self.feat_size = self.max_len*self.sr //self.hop_size
        self.audio_size = self.max_len*self.sr
        self.center = center
        self.overlap = overlap
        self.debug = debug
        self.contiguous = contiguous
        self.contiguous_clip_noise = contiguous_clip_noise

    def set_confidence(self,confidence):
        self.crepe_params.confidence_threshold = confidence

    def process_indices(self, indices: list) -> list:
        # Length in samples.
        max_len = self.max_len * self.sr

        def expand_long(indices_tuple: tuple) -> list:
            if indices_tuple[1] - indices_tuple[0] > max_len:
                ret = [(start, start+max_len) for start in np.arange(indices_tuple[0], indices_tuple[1] - max_len, max_len)]
                ret.append((ret[-1][-1], min(ret[-1][-1] + max_len, indices_tuple[1])))
                return ret
            else:
                return [indices_tuple]

        new_indices = [*map(expand_long, indices)]
        new_indices = functools.reduce(operator.concat, new_indices, [])
        new_indices = [x for x in new_indices if (x[1] - x[0] > self.seq_len * self.sr)]
        return new_indices

    def pad_to_expected_size(self,features,expected_size,pad_value):

        #Pad to next integer division if we are processing a whole file in one go.
        if(self.contiguous == True):
            # Pad up to next integer division
            pad_len = (features.shape[-1] // expected_size + 1)*expected_size - features.shape[-1]
            #print(f'feat len {features.shape[-1]} expected {expected_size} pad {pad_len}')
            features = np.pad(features,(0,pad_len),'constant',constant_values=pad_value)
            return features
        else:
            if(self.debug):
                print("Feat shape {} - expected size: {}".format(features.shape[-1],expected_size))
            if(features.shape[-1] < expected_size):
                pad_len = expected_size - features.shape[-1]
                features = np.pad(features,(0,pad_len),'constant',constant_values=pad_value)
            if(features.shape[-1] > expected_size):
                raise Exception('Expected size is smaller than current value')
        return features


    def extract_f0(self, audio):
        if isinstance(self.crepe_params, dict):
            self.crepe_params = dotdict(self.crepe_params)
        (f0,confidence) = spectral_ops.calc_f0(audio,
                                rate=self.sr,
                                hop_size=self.hop_size,
                                fmin=self.crepe_params.fmin,
                                fmax=self.crepe_params.fmax,
                                model=self.crepe_params.model,
                                batch_size=self.crepe_params.batch_size,
                                device=self.device,
                                center=self.center)

        if confidence.mean() < self.crepe_params.confidence_threshold:
            #print("Low confidence: {}".format(confidence.mean()))
            raise ValueError('Low f0 confidence')

        f0 = self.pad_to_expected_size(f0,
                expected_size = self.feat_size,
                pad_value=0)

        return f0

    def calc_loudness(self,audio):
        if isinstance(self.loudness_params, dict):
            self.loudness_params = dotdict(self.loudness_params)
        loudness = spectral_ops.calc_loudness(audio, rate=self.sr,
                                            n_fft=self.loudness_params.nfft,
                                            hop_size=self.hop_size,
                                            center=self.center,)

        loudness = self.pad_to_expected_size(loudness,
                expected_size = self.feat_size,
                pad_value=-_DB_RANGE)
        return loudness

    # TODO: Add center padding capability here.
    def calc_rms(self,audio):
        if isinstance(self.rms, dict):
            self.rms = dotdict(self.rms)
        rms = spectral_ops.calc_power(audio, frame_size=self.rms.frame_size,
                                        hop_size=self.hop_size,pad_end=True)
        rms = self.pad_to_expected_size(rms,
                expected_size = self.feat_size,
                pad_value=-_DB_RANGE)
        return rms

    def save_data(self, audio, f0, loudness, rms, h5f, counter):
        h5f.create_dataset(f'{counter}_audio', data=audio)
        h5f.create_dataset(f'{counter}_f0', data=f0)
        h5f.create_dataset(f'{counter}_loudness', data=loudness)
        h5f.create_dataset(f'{counter}_rms', data=rms)
        return counter + 1

    def init_h5(self, data_dir):
        return h5py.File(data_dir / f'{self.sr}.h5', 'w')

    def close_h5(self, h5f):
        h5f.close()

    '''
    Main audio processing function
    '''
    def run_on_files(self, data_dir, input_dir, output_dir):
        audio_files = list((input_dir/data_dir).glob('*.wav'))
        output_dir = output_dir / data_dir
        output_dir.mkdir(exist_ok=True)

        # Open container
        h5f = self.init_h5(output_dir)
        counter = 0

        for audio_file in tqdm(audio_files):
            if(self.debug): print("Processing: {}".format(audio_file))

            # load and split files
            data, sr = librosa.load(audio_file.as_posix(), sr=self.sr)
            data = librosa.util.normalize(data) # Peak-normalize audio
            sounds_indices = []
            if(self.contiguous):
                sounds_indices.append([0,len(data)])
            else:
                sounds_indices = librosa.effects.split(data, top_db=self.silence_thresh_dB)
                #print("[DEBUG] Sound indices {}".format(sounds_indices))
                sounds_indices = self.process_indices(sounds_indices)
            if len(sounds_indices) == 0:
                continue


            for indices in sounds_indices:
                audio = data[indices[0]:indices[1]]
                if(self.debug): print("\tIndexes: {} {} - len: {}".format(indices[0],indices[1],indices[1]-indices[0]))

                # Feature retrieval segment

                try: # Only process audio with enough CREPE confidence
                    f0 = self.extract_f0(audio)
                except ValueError:
                    continue

                # Further downsamples the audio back to the other specified sample rates and returns a dictionary.
                loudness = self.calc_loudness(audio)
                rms = self.calc_rms(audio)
                if(self.contiguous):
                    if(self.contiguous_clip_noise):
                        if(self.debug): print("[DEBUG] clipping noise")
                        clip_pos = (f0 > 1900.0)
                        loudness[clip_pos] = -_DB_RANGE
                    audio = self.pad_to_expected_size(audio,f0.shape[0]*self.hop_size,0)

                else:
                    audio = self.pad_to_expected_size(audio,self.audio_size,0)
                if(self.debug): print(f'\t Store block {counter}: f0 : {f0.shape} - loudness : {loudness.shape} - rms {rms.shape} - audio : {audio.shape}')
                counter = self.save_data(audio, f0, loudness, rms, h5f, counter)

        # Finished storing f0 and loudness
        self.close_h5(h5f)


    def run_on_dirs(self, input_dir: Path, output_dir: Path):
        #print("Starting with crepe confidence: {}".format(self.crepe_params.confidence_threshold))
        folders = [x for x in input_dir.glob('./*') if x.is_dir()]
        for folder in tqdm(folders):
            self.run_on_files(folder.name, input_dir, output_dir)


class F0LoudnessRMSPreprocessor():
    """Scales 'f0_hz' and 'loudness_db' features."""
    def __init__(self):
        return

    def run(self,x):
        x['loudness_scaled'] = self.scale_db(x['loudness'])
        x['rms_scaled'] = self.scale_db(x['rms'])
        x['f0_scaled'] = self.scale_f0_hz(x['f0'])
        return x

    def scale_db(self,db):
        """Scales [-DB_RANGE, 0] to [0, 1]."""
        return (db / _DB_RANGE) + 1.0

    def scale_f0_hz(self,f0_hz):
        """Scales [0, Nyquist] Hz to [0, 1.0] MIDI-scaled."""
        return hz_to_midi(f0_hz) / _F0_RANGE


"""
Core functions. 
The code mainly comes from https://github.com/acids-ircam/ddsp_pytorch with minor adaptations.
"""
import torch
import torch.nn as nn
import torch.fft as fft
import numpy as np
import librosa as li
import crepe
from torchcrepeV2 import TorchCrepePredictor
import math
import matplotlib.pyplot as plt
import yaml 


with open("syntheon/inferencer/vital/config.yaml", 'r') as stream:
    config = yaml.safe_load(stream)

device = config["device"]
if device == "cuda":
    crepe_predictor = TorchCrepePredictor()
else:
    crepe_predictor = TorchCrepePredictor(device="cpu")


def safe_log(x):
    return torch.log(x + 1e-7)


@torch.no_grad()
def mean_std_loudness(dataset):
    mean = 0
    std = 0
    n = 0
    for _, _, l in dataset:
        n += 1
        mean += (l.mean().item() - mean) / n
        std += (l.std().item() - std) / n
    return mean, std


def multiscale_fft(signal, scales, overlap):
    stfts = []
    for s in scales:
        S = torch.stft(
            signal,
            s,
            int(s * (1 - overlap)),
            s,
            torch.hann_window(s).to(signal),
            True,
            normalized=True,
            return_complex=True,
        ).abs()
        stfts.append(S)
    return stfts


def resample(x, factor: int):
    batch, frame, channel = x.shape
    x = x.permute(0, 2, 1).reshape(batch * channel, 1, frame)

    window = torch.hann_window(
        factor * 2,
        dtype=x.dtype,
        device=x.device,
    ).reshape(1, 1, -1)
    y = torch.zeros(x.shape[0], x.shape[1], factor * x.shape[2]).to(x)
    y[..., ::factor] = x
    y[..., -1:] = x[..., -1:]
    y = torch.nn.functional.pad(y, [factor, factor])
    y = torch.nn.functional.conv1d(y, window)[..., :-1]

    y = y.reshape(batch, channel, factor * frame).permute(0, 2, 1)

    return y


def upsample(signal, factor, preferred_size=None, mode="nearest"):
    signal = signal.permute(0, 2, 1)
    if preferred_size is not None:
        signal = nn.functional.interpolate(signal, size=preferred_size, mode=mode)
    else:
        signal = nn.functional.interpolate(signal, size=signal.shape[-1] * factor, mode=mode)
    return signal.permute(0, 2, 1)


def remove_above_nyquist(amplitudes, pitch, sampling_rate):
    n_harm = amplitudes.shape[-1]
    pitches = pitch * torch.arange(1, n_harm + 1).to(pitch)
    aa = (pitches < sampling_rate / 2).float() + 1e-4
    return amplitudes * aa


def scale_function(x):
    return 2 * torch.sigmoid(x)**(math.log(10)) + 1e-7


def amplitude_to_db(amplitude):
    amin = 1e-20  # Avoid log(0) instabilities.
    db = torch.log10(torch.clamp(amplitude, min=amin))
    db *= 20.0
    return db


def extract_loudness(audio, sampling_rate, block_size=None, n_fft=2048, frame_rate=None):
    assert (block_size is None) != (frame_rate is None), "Specify exactly one of block_size or frame_rate"

    if frame_rate is not None:
        block_size = sampling_rate // frame_rate
    else:
        frame_rate = int(sampling_rate / block_size)

    if sampling_rate % frame_rate != 0:
        raise ValueError(
            'frame_rate: {} must evenly divide sample_rate: {}.'
            'For default frame_rate: 250Hz, suggested sample_rate: 16kHz or 48kHz'
            .format(frame_rate, sampling_rate))

    if isinstance(audio, np.ndarray):
        audio = torch.tensor(audio)

    # Temporarily a batch dimension for single examples.
    is_1d = (len(audio.shape) == 1)
    audio = audio[None, :] if is_1d else audio

    # Take STFT.
    overlap = 1 - block_size / n_fft
    amplitude = torch.stft(audio, n_fft=n_fft, hop_length=block_size, center=True, pad_mode='reflect', return_complex=True).abs()
    amplitude = amplitude[:, :, :-1]
    
    # Compute power.
    power_db = amplitude_to_db(amplitude)

    # Perceptual weighting.
    frequencies = li.fft_frequencies(sr=sampling_rate, n_fft=n_fft)
    a_weighting = li.A_weighting(frequencies)[None,:,None]
    loudness = power_db + a_weighting

    loudness = torch.mean(torch.pow(10, loudness / 10.0), axis=1)
    loudness = 10.0 * torch.log10(torch.clamp(loudness, min=1e-20))

    # Remove temporary batch dimension.
    loudness = loudness[0] if is_1d else loudness
    loudness = loudness.numpy()

    return loudness


def extract_pitch(signal, sampling_rate, block_size, model_capacity="full"):
    length = signal.shape[-1] // block_size
    if device == "cpu":
        f0 = crepe.predict(
            signal,
            sampling_rate,
            step_size=int(1000 * block_size / sampling_rate),
            verbose=1,
            center=True,
            viterbi=True,
            model_capacity="full"
        )
        f0 = f0[1].reshape(-1)[:-1]
    else:
        f0 = crepe_predictor.predict(
            signal,
            sampling_rate
        )

    if f0.shape[-1] != length:
        f0 = np.interp(
            np.linspace(0, 1, length, endpoint=False),
            np.linspace(0, 1, f0.shape[-1], endpoint=False),
            f0,
        )

    return f0


def mlp(in_size, hidden_size, n_layers):
    channels = [in_size] + (n_layers) * [hidden_size]
    net = []
    for i in range(n_layers):
        net.append(nn.Linear(channels[i], channels[i + 1]))
        net.append(nn.LayerNorm(channels[i + 1]))
        net.append(nn.LeakyReLU())
    return nn.Sequential(*net)


def gru(n_input, hidden_size):
    return nn.GRU(n_input * hidden_size, hidden_size, batch_first=True)


def harmonic_synth(pitch, amplitudes, sampling_rate):
    n_harmonic = amplitudes.shape[-1]
    omega = torch.cumsum(2 * math.pi * pitch / sampling_rate, 1)
    omegas = omega * torch.arange(1, n_harmonic + 1).to(omega)

    signal = (torch.sin(omegas) * amplitudes).sum(-1, keepdim=True)
    return signal


def amp_to_impulse_response(amp, target_size):
    amp = torch.stack([amp, torch.zeros_like(amp)], -1)
    amp = torch.view_as_complex(amp)
    amp = fft.irfft(amp)

    filter_size = amp.shape[-1]

    amp = torch.roll(amp, filter_size // 2, -1)
    win = torch.hann_window(filter_size, dtype=amp.dtype, device=amp.device)

    amp = amp * win

    amp = nn.functional.pad(amp, (0, int(target_size) - int(filter_size)))
    amp = torch.roll(amp, -filter_size // 2, -1)

    return amp


def fft_convolve(signal, kernel):
    signal = nn.functional.pad(signal, (0, signal.shape[-1]))
    kernel = nn.functional.pad(kernel, (kernel.shape[-1], 0))

    output = fft.irfft(fft.rfft(signal) * fft.rfft(kernel))
    output = output[..., output.shape[-1] // 2:]

    return output


def get_scheduler(len_dataset, start_lr, stop_lr, length):
    def schedule(epoch):
        step = epoch * len_dataset
        if step < length:
            t = step / length
            return start_lr * (1 - t) + stop_lr * t
        else:
            return stop_lr

    return schedule


import torch
import torch.nn as nn
import torch.fft as fft
import numpy as np
import librosa as li
import math

_DB_RANGE = 80.0 #Min loudness
_REF_DB = 20.7  # White noise, amplitude=1.0, n_fft=2048
_F0_RANGE = 127

def safe_log(x,eps=1e-7):
    eps = torch.tensor(eps)
    return torch.log(x + eps)

def safe_divide(numerator, denominator, eps=1e-7):
    """Avoid dividing by zero by adding a small epsilon."""
    eps = torch.tensor(eps)
    safe_denominator = torch.where(denominator == 0.0, eps, denominator)
    return numerator / safe_denominator

def logb(x, base=2.0, eps=1e-5):
    """Logarithm with base as an argument."""
    return safe_divide(safe_log(x, eps), safe_log(base, eps), eps)

def hz_to_midi(frequencies):
    """Torch-compatible hz_to_midi function."""
    notes = 12.0 * (logb(frequencies, 2.0) - logb(440.0, 2.0)) + 69.0
    notes = torch.where(torch.le(frequencies, torch.zeros(1).to(frequencies)),
                        torch.zeros(1).to(frequencies), notes)
    return notes


@torch.no_grad()
def cumsum_nd(in_tensor,wrap_value=None):
    '''
    cumsum_nd() : cummulative sum - non differentiable and with wrap value.

    The problem with cumsum: when we work with phase tensors that are too large
    (i.e. more than a few tenths of seconds) cumsum gets to accumulate steps
    over a very large window, and it seems the float point variable loses precision.

    This workaround computes the accumulation step by step, resetting the
    accumulator in order for it to avoid to lose precision.

    NOTE: This implementation is very slow, and can't be used during training,
    only for final audio rendering on the test set.

    Assumes a tensor format used for audio rendering. [batch,len,1]

    NOTE:  Non integer frequency ratios do not work using current synthesis approach,
    because we render a common phase (wrapped using cumsum_nd) and then we multiply it
    by the frequency ratio. This introduces a misalignment if we multiply the wrapped phase
    by a non-integer frequency ratio.

    TODO: implement an efficient vectorial cumsum with wrapping we can use to accumulate
          phases from all oscillators separately
    '''
    print("[WARNING] Using non differentiable cumsum. Non-integer frequency ratios wont render well.")
    input_len = in_tensor.size()[1]
    nb = in_tensor.size()[0]
    acc = torch.zeros([nb,1,1])
    out_tensor = torch.zeros([nb,input_len,1])
    #print("in size{} - out size{}".format(in_tensor.size(),out_tensor.size()))
    for i in range(input_len):
        acc += in_tensor[:,i,0]
        if(wrap_value is not None):
            acc = acc - (acc > wrap_value)*wrap_value
        out_tensor[:,i,0] = acc
    return out_tensor



@torch.no_grad()
def mean_std_loudness(dataset):
    mean = 0
    std = 0
    n = 0
    for _, _, l in dataset:
        n += 1
        mean += (l.mean().item() - mean) / n
        std += (l.std().item() - std) / n
    return mean, std


def multiscale_fft(signal, scales, overlap):
    stfts = []
    for s in scales:
        S = torch.stft(
            signal,
            s,
            int(s * (1 - overlap)),
            s,
            torch.hann_window(s).to(signal),
            True,
            normalized=True,
            return_complex=True,
        ).abs()
        stfts.append(S)
    return stfts


def resample(x, factor: int):
    batch, frame, channel = x.shape
    x = x.permute(0, 2, 1).reshape(batch * channel, 1, frame)

    window = torch.hann_window(
        factor * 2,
        dtype=x.dtype,
        device=x.device,
    ).reshape(1, 1, -1)
    y = torch.zeros(x.shape[0], x.shape[1], factor * x.shape[2]).to(x)
    y[..., ::factor] = x
    y[..., -1:] = x[..., -1:]
    y = torch.nn.functional.pad(y, [factor, factor])
    y = torch.nn.functional.conv1d(y, window)[..., :-1]

    y = y.reshape(batch, channel, factor * frame).permute(0, 2, 1)

    return y


def upsample(signal, factor,mode='nearest'):
    signal = signal.permute(0, 2, 1)
    signal = nn.functional.interpolate(signal, size=signal.shape[-1] * factor,mode=mode)
    return signal.permute(0, 2, 1)


def extract_loudness(signal, sampling_rate, block_size, n_fft=2048):
    S = li.stft(
        signal,
        n_fft=n_fft,
        hop_length=block_size,
        win_length=n_fft,
        center=True,
    )
    S = np.log(abs(S) + 1e-7)
    f = li.fft_frequencies(sampling_rate, n_fft)
    a_weight = li.A_weighting(f)

    S = S + a_weight.reshape(-1, 1)

    S = np.mean(S, 0)[..., :-1]

    return S



def get_mlp(in_size, hidden_size, n_layers):
    channels = [in_size] + (n_layers) * [hidden_size]
    net = []
    for i in range(n_layers):
        net.append(nn.Linear(channels[i], channels[i + 1]))
        net.append(nn.LayerNorm(channels[i + 1]))
        net.append(nn.LeakyReLU())
    return nn.Sequential(*net)


def get_gru(n_input, hidden_size):
    return nn.GRU(n_input * hidden_size, hidden_size, batch_first=True)


def amp_to_impulse_response(amp, target_size):
    amp = torch.stack([amp, torch.zeros_like(amp)], -1)
    amp = torch.view_as_complex(amp)
    amp = fft.irfft(amp)

    filter_size = amp.shape[-1]

    amp = torch.roll(amp, filter_size // 2, -1)
    win = torch.hann_window(filter_size, dtype=amp.dtype, device=amp.device)

    amp = amp * win

    amp = nn.functional.pad(amp, (0, int(target_size) - int(filter_size)))
    amp = torch.roll(amp, -filter_size // 2, -1)

    return amp


def fft_convolve(signal, kernel):
    signal = nn.functional.pad(signal, (0, signal.shape[-1]))
    kernel = nn.functional.pad(kernel, (kernel.shape[-1], 0))

    output = fft.irfft(fft.rfft(signal) * fft.rfft(kernel))
    output = output[..., output.shape[-1] // 2:]

    return output


def harmonic_synth(pitch, amplitudes, sampling_rate,use_safe_cumsum=False):

    if(use_safe_cumsum==True):
        omega = cumsum_nd(2 * np.pi * pitch / sampling_rate, 2*np.pi)
    else:
        omega = torch.cumsum(2 * np.pi * pitch / sampling_rate, 1)

    n_harmonic = amplitudes.shape[-1]
    omegas = omega * torch.arange(1, n_harmonic + 1).to(omega)
    signal = (torch.sin(omegas) * amplitudes).sum(-1, keepdim=True)
    return signal

OP6=5
OP5=4
OP4=3
OP3=2
OP2=1
OP1=0

def fm_2stack2(pitch, ol, fr, sampling_rate,max_ol,use_safe_cumsum=False):

    if(use_safe_cumsum==True):
        omega = cumsum_nd(2 * np.pi * pitch / sampling_rate, 2*np.pi)
    else:
        omega = torch.cumsum(2 * np.pi * pitch / sampling_rate, 1)

    # Torch unsqueeze with dim -1 adds a new dimension at the end of ol to match phases.

    op4_phase =  fr[OP4] * omega
    op4_output = torch.unsqueeze(ol[:,:,OP4], dim=-1) * torch.sin(op4_phase)

    op3_phase =  fr[OP3] * omega + 2 * np.pi * op4_output
    op3_output = torch.unsqueeze(ol[:,:,OP3], dim=-1) * torch.sin(op3_phase) # output of stack of 2

    op2_phase =  fr[OP2] * omega
    op2_output = torch.unsqueeze(ol[:,:,OP2], dim=-1) * torch.sin(op2_phase)

    op1_phase =  fr[OP1] * omega + 2 * np.pi * op2_output
    op1_output = torch.unsqueeze(ol[:,:,OP1], dim=-1) * torch.sin(op1_phase) # output of stack of 2

    return (op3_output + op1_output)/max_ol

def fm_1stack2(pitch, ol, fr, sampling_rate,max_ol,use_safe_cumsum=False):

    if(use_safe_cumsum==True):
        omega = cumsum_nd(2 * np.pi * pitch / sampling_rate, 2*np.pi)
    else:
        omega = torch.cumsum(2 * np.pi * pitch / sampling_rate, 1)

    # Torch unsqueeze with dim -1 adds a new dimension at the end of ol to match phases.

    op2_phase =  fr[OP2] * omega
    op2_output = torch.unsqueeze(ol[:,:,OP2], dim=-1) * torch.sin(op2_phase)

    op1_phase =  fr[OP1] * omega + 2 * np.pi * op2_output
    op1_output = torch.unsqueeze(ol[:,:,OP1], dim=-1) * torch.sin(op1_phase) # output of stack of 2

    return op1_output/max_ol


def fm_1stack4(pitch, ol, fr, sampling_rate,max_ol,use_safe_cumsum=False):

    if(use_safe_cumsum==True):
        omega = cumsum_nd(2 * np.pi * pitch / sampling_rate, 2*np.pi)
    else:
        omega = torch.cumsum(2 * np.pi * pitch / sampling_rate, 1)

    # Torch unsqueeze with dim -1 adds a new dimension at the end of ol to match phases.

    op4_phase =  fr[OP4] * omega
    op4_output = torch.unsqueeze(ol[:,:,OP4], dim=-1) * torch.sin(op4_phase)

    op3_phase =  fr[OP3] * omega + 2 * np.pi * op4_output
    op3_output = torch.unsqueeze(ol[:,:,OP3], dim=-1) * torch.sin(op3_phase) # output of stack of 4

    op2_phase =  fr[OP2] * omega + 2 * np.pi * op3_output
    op2_output = torch.unsqueeze(ol[:,:,OP2], dim=-1) * torch.sin(op2_phase)

    op1_phase =  fr[OP1] * omega + 2 * np.pi * op2_output
    op1_output = torch.unsqueeze(ol[:,:,OP1], dim=-1) * torch.sin(op1_phase) # output of stack of 2

    return op1_output/max_ol


'''
Ablated Brass FM Synth - with phase wrapping (it does not change behaviour)
     OP4->OP3->|
          OP2->|->OP1->out

'''
def fm_ablbrass_synth(pitch, ol, fr, sampling_rate,max_ol,use_safe_cumsum=False):

    if(use_safe_cumsum==True):
        omega = cumsum_nd(2 * np.pi * pitch / sampling_rate, 2*np.pi)
    else:
        omega = torch.cumsum(2 * np.pi * pitch / sampling_rate, 1)

    # Torch unsqueeze with dim -1 adds a new dimension at the end of ol to match phases.

    op4_phase =  fr[OP4] * omega
    op4_output = torch.unsqueeze(ol[:,:,OP4], dim=-1) * torch.sin(op4_phase % (2*np.pi))

    op3_phase =  fr[OP3] * omega + 2 * np.pi * op4_output
    op3_output = torch.unsqueeze(ol[:,:,OP3], dim=-1) * torch.sin(op3_phase % (2*np.pi)) # output of stack of 2

    op2_phase =  fr[OP2] * omega
    op2_output = torch.unsqueeze(ol[:,:,OP2], dim=-1) * torch.sin(op2_phase % (2*np.pi)) # output stack of 1

    op1_phase =  fr[OP1] * omega + 2 * np.pi * (op2_output + op3_output)
    op1_output = torch.unsqueeze(ol[:,:,OP1], dim=-1) * torch.sin(op1_phase % (2*np.pi)) # global carrier

    return op1_output/max_ol

'''
String FM Synth - with phase wrapping (it does not change behaviour)
PATCH NAME: STRINGS 1
OP6->OP5->OP4->OP3 |
       (R)OP2->OP1 |->out
'''
def fm_string_synth(pitch, ol, fr, sampling_rate,max_ol,use_safe_cumsum=False):

    if(use_safe_cumsum==True):
        omega = cumsum_nd(2 * np.pi * pitch / sampling_rate, 2*np.pi)
    else:
        omega = torch.cumsum(2 * np.pi * pitch / sampling_rate, 1)

    # Torch unsqueeze with dim -1 adds a new dimension at the end of ol to match phases.
    op6_phase =  fr[OP6] * omega
    op6_output = torch.unsqueeze(ol[:,:,OP6], dim=-1) * torch.sin(op6_phase % (2*np.pi))

    op5_phase =  fr[OP5] * omega + 2 * np.pi * op6_output
    op5_output = torch.unsqueeze(ol[:,:,OP5], dim=-1)*torch.sin(op5_phase % (2*np.pi))

    op4_phase =  fr[OP4] * omega + 2 * np.pi * op5_output
    op4_output = torch.unsqueeze(ol[:,:,OP4], dim=-1) * torch.sin(op4_phase % (2*np.pi))

    op3_phase =  fr[OP3] * omega + 2 * np.pi * op4_output
    op3_output = torch.unsqueeze(ol[:,:,OP3], dim=-1) * torch.sin(op3_phase % (2*np.pi)) # output of stack of 4

    op2_phase =  fr[OP2] * omega
    op2_output = torch.unsqueeze(ol[:,:,OP2], dim=-1) * torch.sin(op2_phase % (2*np.pi))

    op1_phase =  fr[OP1] * omega + 2 * np.pi * op2_output
    op1_output = torch.unsqueeze(ol[:,:,OP1], dim=-1) * torch.sin(op1_phase % (2*np.pi)) # output of stack of 2

    return (op3_output + op1_output)/max_ol

'''
Flute FM Synth - with phase wrapping (it does not change behaviour)
PATCH NAME: FLUTE 1
(R)OP6->OP5->|
   OP4->OP3->|
        OP2->|->OP1->out
'''
def fm_flute_synth(pitch, ol, fr, sampling_rate,max_ol,use_safe_cumsum=False):

    if(use_safe_cumsum==True):
        omega = cumsum_nd(2 * np.pi * pitch / sampling_rate, 2*np.pi)
    else:
        omega = torch.cumsum(2 * np.pi * pitch / sampling_rate, 1)

    # Torch unsqueeze with dim -1 adds a new dimension at the end of ol to match phases.
    op6_phase =  fr[OP6] * omega
    op6_output = torch.unsqueeze(ol[:,:,OP6], dim=-1) * torch.sin(op6_phase % (2*np.pi))

    op5_phase =  fr[OP5] * omega + 2 * np.pi * op6_output
    op5_output = torch.unsqueeze(ol[:,:,OP5], dim=-1)*torch.sin(op5_phase % (2*np.pi)) # output of stack of 2

    op4_phase =  fr[OP4] * omega
    op4_output = torch.unsqueeze(ol[:,:,OP4], dim=-1) * torch.sin(op4_phase % (2*np.pi))

    op3_phase =  fr[OP3] * omega + 2 * np.pi * op4_output
    op3_output = torch.unsqueeze(ol[:,:,OP3], dim=-1) * torch.sin(op3_phase % (2*np.pi)) # output of stack of 2

    op2_phase =  fr[OP2] * omega
    op2_output = torch.unsqueeze(ol[:,:,OP2], dim=-1) * torch.sin(op2_phase % (2*np.pi)) # output stack of 1

    op1_phase =  fr[OP1] * omega + 2 * np.pi * (op2_output + op3_output + op5_output)
    op1_output = torch.unsqueeze(ol[:,:,OP1], dim=-1) * torch.sin(op1_phase % (2*np.pi)) # carrier

    return op1_output/max_ol

'''
Brass FM Synth - with phase wrapping (it does not change behaviour)
PATCH NAME: BRASS 3
OP6->OP5->OP4->|
       (R)OP3->|
          OP2->|->OP1->out
'''
def fm_brass_synth(pitch, ol, fr, sampling_rate,max_ol,use_safe_cumsum=False):

    if(use_safe_cumsum==True):
        omega = cumsum_nd(2 * np.pi * pitch / sampling_rate, 2*np.pi)
    else:
        omega = torch.cumsum(2 * np.pi * pitch / sampling_rate, 1)

    # Torch unsqueeze with dim -1 adds a new dimension at the end of ol to match phases.
    op6_phase =  fr[OP6] * omega
    op6_output = torch.unsqueeze(ol[:,:,OP6], dim=-1) * torch.sin(op6_phase % (2*np.pi))

    op5_phase =  fr[OP5] * omega + 2 * np.pi * op6_output
    op5_output = torch.unsqueeze(ol[:,:,OP5], dim=-1)*torch.sin(op5_phase % (2*np.pi))

    op4_phase =  fr[OP4] * omega + 2 * np.pi * op5_output
    op4_output = torch.unsqueeze(ol[:,:,OP4], dim=-1) * torch.sin(op4_phase % (2*np.pi)) # output of stack of 3

    op3_phase =  fr[OP3] * omega
    op3_output = torch.unsqueeze(ol[:,:,OP3], dim=-1) * torch.sin(op3_phase % (2*np.pi)) # output of stack of 1

    op2_phase =  fr[OP2] * omega
    op2_output = torch.unsqueeze(ol[:,:,OP2], dim=-1) * torch.sin(op2_phase % (2*np.pi)) # output stack of 1

    op1_phase =  fr[OP1] * omega + 2 * np.pi * (op2_output + op3_output + op4_output)
    op1_output = torch.unsqueeze(ol[:,:,OP1], dim=-1) * torch.sin(op1_phase % (2*np.pi)) # carrier

    return op1_output/max_ol



"""
Function APIs to be called externally.
"""
from .converter.dexed.dexed_converter import DexedConverter
from .inferencer.dexed.dexed_inferencer import DexedInferencer
from .converter.vital.vital_converter import VitalConverter
from .inferencer.vital.vital_inferencer import VitalInferencer


obj_dict = {
    "dexed": {
        "converter": DexedConverter,
        "inferencer": DexedInferencer
    },
    "vital": {
        "converter": VitalConverter,
        "inferencer": VitalInferencer
    }
}

def infer_params(input_audio_name, synth_name, enable_eval=False):
    if synth_name not in obj_dict:
        raise ValueError("Synth name {} not available for parameter inference".format(synth_name))
    
    inferencer = obj_dict[synth_name]["inferencer"](device="cpu")
    params, eval_dict = inferencer.convert(input_audio_name, enable_eval=enable_eval)

    converter = obj_dict[synth_name]["converter"]()
    converter.dict = params
    output_fname = "{}_output.syx".format(synth_name)
    converter.parseToPluginFile(output_fname)

    return output_fname, eval_dict


import ddx7.core as core

class F0LoudnessRMSPreprocessor():
    """Scales 'f0_hz' and 'loudness_db' features."""
    def __init__(self):
        return

    def run(self,x):
        x['loudness_scaled'] = self.scale_db(x['loudness'])
        x['rms_scaled'] = self.scale_db(x['rms'])
        x['f0_scaled'] = self.scale_f0_hz(x['f0'])
        return x

    def scale_db(self,db):
        """Scales [-DB_RANGE, 0] to [0, 1]."""
        return (db / core._DB_RANGE) + 1.0

    def scale_f0_hz(self,f0_hz):
        """Scales [0, Nyquist] Hz to [0, 1.0] MIDI-scaled."""
        return core.hz_to_midi(f0_hz) / core._F0_RANGE


"""
Common class for pitch extraction.
"""

import numpy as np
import crepe
from torchcrepeV2 import TorchCrepePredictor
import yaml 

with open("syntheon/inferencer/vital/config.yaml", 'r') as stream:
    config = yaml.safe_load(stream)

device = config["device"]
if device == "cuda":
    crepe_predictor = TorchCrepePredictor()
else:
    crepe_predictor = TorchCrepePredictor(device="cpu")


def extract_pitch(signal, sampling_rate, block_size, model_capacity="full"):
    length = signal.shape[-1] // block_size
    if device == "cpu":
        f0 = crepe.predict(
            signal,
            sampling_rate,
            step_size=int(1000 * block_size / sampling_rate),
            verbose=1,
            center=True,
            viterbi=True,
            model_capacity="full"
        )
        f0 = f0[1].reshape(-1)[:-1]
    else:
        f0 = crepe_predictor.predict(
            signal,
            sampling_rate
        )

    if f0.shape[-1] != length:
        f0 = np.interp(
            np.linspace(0, 1, length, endpoint=False),
            np.linspace(0, 1, f0.shape[-1], endpoint=False),
            f0,
        )

    return f0


import torch
import torch.nn as nn
from syntheon.inferencer.dexed.models.ddx7.core import get_gru, get_mlp
import torch.nn.functional as F

'''
Wrapper class for either HpN or DDX7
'''
class DDSP_Decoder(nn.Module):
    def __init__(self,decoder,synth):
        super().__init__()
        net = []
        net.append(decoder)
        net.append(synth)
        self.net = nn.Sequential(*net)

    def forward(self,x):
        return self.net(x)

    def get_sr(self):
        return self.net[1].sample_rate

    def enable_cumsum_nd(self):
        self.net[1].use_cumsum_nd=True

    def get_params(self,param):
        if(param == 'reverb_decay'):
            return self.net[1].reverb.decay.item()
        if(param == 'reverb_wet'):
            return self.net[1].reverb.wet.item()

'''
GRU-Based decoder for HpN Baseline
'''
class RnnFCDecoder(nn.Module):
    def __init__(self, hidden_size=512, sample_rate=16000,
                 input_keys=None,input_sizes=[1,1,16],
                 output_keys=['amplitude','harmonic_distribution','noise_bands'],
                 output_sizes=[1,100,65]):
        super().__init__()
        self.input_keys = input_keys
        self.input_sizes = input_sizes
        n_keys = len(input_keys)
        # Generate MLPs of size: in_size: 1 ; n_layers = 3 (with layer normalization and leaky relu)
        if(n_keys == 2):
            self.in_mlps = nn.ModuleList([get_mlp(input_sizes[0], hidden_size, 3),
                                          get_mlp(input_sizes[1], hidden_size, 3)])
        elif(n_keys == 3):
            self.in_mlps = nn.ModuleList([get_mlp(input_sizes[0], hidden_size, 3),
                                          get_mlp(input_sizes[1], hidden_size, 3),
                                          get_mlp(input_sizes[2], hidden_size, 3)])
        else:
            raise ValueError("Expected 2 or 3 input keys. got: {}".format(input_keys))

        #Generate GRU: input_size = n_keys * hidden_size ; n_layers = 1 (that's the default config)
        self.gru = get_gru(n_keys, hidden_size)

        #Generate output MLP: in_size: hidden_size + 2 ; n_layers = 3
        self.out_mlp = get_mlp(hidden_size + 2, hidden_size, 3)

        self.proj_matrices = []
        self.output_keys = output_keys
        self.output_sizes = output_sizes
        for v,k in enumerate(output_keys):
            self.proj_matrices.append(nn.Linear(hidden_size,output_sizes[v]))

        self.proj_matrices = nn.ModuleList(self.proj_matrices)
        self.sample_rate = sample_rate

    def forward(self, x):
        # Run pitch and loudness and z (if available) inputs through the respectives input MLPs.
        # Then, concatenate the outputs in a flat vector.

        # Run through input_keys and load inputs accordingly
        hidden = torch.cat([self.in_mlps[v](x[k]) for v,k in enumerate(self.input_keys)],-1)

        # Run the flattened vector through the GRU.
        # The GRU predicts the embedding.
        # Then, concatenate the embedding with the disentangled parameters of pitch and loudness (nhid+2 size vector)
        hidden = torch.cat([self.gru(hidden)[0], x['f0_scaled'], x['loudness_scaled']], -1)
        # Run the embedding through the output MLP to obtain a 512-sized output vector.
        hidden = self.out_mlp(hidden)


        # Run embedding through a projection_matrix to get outputs
        controls = {}
        for v,k in enumerate(self.output_keys):
            controls[k] = self.proj_matrices[v](hidden)

        controls['f0_hz'] = x['f0']

        return controls

'''
TCN-Based decoder for DDX7
'''
class TCNFMDecoder(nn.Module):
    '''
    FM Decoder with sigmoid output
    '''
    def __init__(self,n_blocks=2,hidden_channels=64,out_channels=6,
                kernel_size=3,dilation_base=2,apply_padding=True,
                deploy_residual=False,
                input_keys=None,z_size=None,
                output_complete_controls=True):
        super().__init__()

        # Store receptive field
        dilation_factor = (dilation_base**n_blocks-1)/(dilation_base-1)
        self.receptive_field = 1 + 2*(kernel_size-1)*dilation_factor
        print("[INFO] TCNFNDecoder - receptive field is: {}".format(self.receptive_field))

        self.input_keys = input_keys
        n_keys = len(input_keys)
        self.output_complete_controls = output_complete_controls

        if(n_keys == 2):
            in_channels = 2
        elif(n_keys == 3):
            in_channels = 2 + z_size
        else:
            raise ValueError("Expected 2 or 3 input keys. got: {}".format(input_keys))

        base = 0
        net = []

        net.append(TCN_block(in_channels,hidden_channels,hidden_channels,kernel_size,
            dilation=dilation_base**base,apply_padding=apply_padding,
            deploy_residual=deploy_residual))
        if(n_blocks>2):
            for i in range(n_blocks-2):
                base += 1
                net.append(TCN_block(hidden_channels,hidden_channels,hidden_channels,
                    kernel_size,dilation=dilation_base**base,apply_padding=apply_padding))

        base += 1
        net.append(TCN_block(hidden_channels,hidden_channels,out_channels,kernel_size,
            dilation=dilation_base**base,apply_padding=apply_padding,
            deploy_residual=deploy_residual,last_block=True))

        self.net = nn.Sequential(*net)

    def forward(self,x):
        # Reshape features to follow Conv1d convention (nb,ch,seq_Len)
        conditioning = torch.cat([x[k] for v,k in enumerate(self.input_keys)],-1).permute([0,-1,-2])

        ol = self.net(conditioning)
        ol = ol.permute([0,-1,-2])
        if self.output_complete_controls is True:
            synth_params = {
                'f0_hz': x['f0'], #In Hz
                'ol': ol
                }
        else:
            synth_params = ol
        return synth_params

class TCN_block(nn.Module):
    '''
    TCN Block
    '''
    def __init__(self,in_channels,hidden_channels,out_channels,
                kernel_size,stride=1,dilation=1,apply_padding=True,
                last_block=False,deploy_residual=False):
        super().__init__()
        block = []
        cnv1 = CausalConv1d(in_channels,hidden_channels,kernel_size,
            stride=stride,dilation=dilation,apply_padding=apply_padding)
        block.append(torch.nn.utils.weight_norm( cnv1 ) )
        block.append(nn.ReLU())
        block.append(nn.Dropout())

        cnv2 = CausalConv1d(hidden_channels,out_channels,kernel_size,
            stride=stride,dilation=dilation,apply_padding=apply_padding)
        block.append(torch.nn.utils.weight_norm( cnv2 ) )
        if(last_block == False):
            block.append(nn.ReLU())
            block.append(nn.Dropout())

        self.block = nn.Sequential(*block)
        self.residual = None
        if(deploy_residual):
            if(apply_padding):
                self.residual = nn.Conv1d(in_channels,out_channels,1,padding = 0,stride=stride)
            else:
                raise ValueError("Residual connection is only possible when padding is enabled.")

    def forward(self,data):
        block_out = self.block(data)
        if(self.residual is not None):
            residual = self.residual(data)
            block_out = block_out + residual
        return block_out


class CausalConv1d(torch.nn.Conv1d):
    '''
    Basic layer for implementing a TCN
    '''
    def __init__(self,
                 in_channels,
                 out_channels,
                 kernel_size,
                 stride=1,
                 dilation=1,
                 groups=1,
                 bias=True,
                 apply_padding=True):

        super(CausalConv1d, self).__init__(
            in_channels,
            out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=0,
            dilation=dilation,
            groups=groups,
            bias=bias)

        self.apply_padding = apply_padding
        self.__padding = dilation*(kernel_size - 1)

    def forward(self, input):
        # Apply left padding using torch.nn.functional and then compute conv.
        if(self.apply_padding):
            return super(CausalConv1d, self).forward(F.pad(input, (self.__padding, 0)))
        else:
            return super(CausalConv1d, self).forward(input)



"""
Differentiable wavetable synthesis component.
"""
import torch
from torch import nn
import numpy as np
from syntheon.inferencer.vital.models.utils import *
from tqdm import tqdm
import soundfile as sf
import matplotlib.pyplot as plt
from syntheon.inferencer.vital.models.core import upsample
import random
import time
from syntheon.inferencer.vital.models.adsr_envelope import *


def wavetable_osc(wavetable, freq, sr):
    """
    General wavetable synthesis oscilator.
    wavetable: (wavetable_len,)
    freq: (batch_size, dur * sr)
    sr: const
    """
    freq = freq.squeeze()
    increment = freq / sr * wavetable.shape[0]
    index = torch.cumsum(increment, dim=1) - increment[0]
    index = index % wavetable.shape[0]

    # uses linear interpolation implementation
    index_low = torch.floor(index.clone())
    index_high = torch.ceil(index.clone())
    alpha = index - index_low
    index_low = index_low.long()
    index_high = index_high.long()

    output = wavetable[index_low] + alpha * (wavetable[index_high % wavetable.shape[0]] - wavetable[index_low])
        
    return output


def wavetable_osc_v2(wavetable, freq, sr):
    """
    General wavetable synthesis oscilator, wavetable per item in batch
    wavetable: (batch_size, wavetable_len,)
    freq: (batch_size, dur * sr)
    sr: const
    """
    freq = freq.squeeze()
    increment = freq / sr * wavetable.shape[1]
    index = torch.cumsum(increment, dim=1) - increment[1]
    index = index % wavetable.shape[1]

    # uses linear interpolation implementation
    index_low = torch.floor(index.clone())
    index_high = torch.ceil(index.clone())
    alpha = index - index_low
    index_low = index_low.long()
    index_high = index_high.long()

    batch_size = wavetable.shape[0]
    output = []

    # TODO: do for loop for now, think any ways to parallelize this (einsum?)
    for bs in range(batch_size):
        wt, idx_l, idx_h, alp = wavetable[bs], index_low[bs].unsqueeze(0), index_high[bs].unsqueeze(0), alpha[bs].unsqueeze(0)
        signal = wt[idx_l] + alp * (wt[idx_h % wt.shape[0]] - wt[idx_l])
        output.append(signal)
    
    output = torch.cat(output, dim=0)
    return output


def generate_wavetable(length, f, cycle=1, phase=0):
    """
    Generate a wavetable of specified length using 
    function f(x) where x is phase.
    Period of f is assumed to be 2 pi.
    """
    wavetable = np.zeros((length,), dtype=np.float32)
    for i in range(length):
        wavetable[i] = f(cycle * 2 * np.pi * i / length + 2 * phase * np.pi)
    return torch.tensor(wavetable)


class WavetableSynth(nn.Module):
    def __init__(self,
                 wavetables=None,
                 n_wavetables=64,
                 wavetable_len=512,
                 sr=44100,
                 duration_secs=3,
                 block_size=160,
                 is_initial_wt_trainable=True):
        super(WavetableSynth, self).__init__()
        if wavetables is None: 
            self.wavetables = []
            for _ in range(n_wavetables):
                cur = nn.Parameter(torch.empty(wavetable_len).normal_(mean=0, std=0.01))
                self.wavetables.append(cur)

            self.wavetables = nn.ParameterList(self.wavetables)

            for idx, wt in enumerate(self.wavetables):
                # following the paper, initialize f0-f3 wavetables and disable backprop
                # if idx == 0:
                #     wt.data = generate_wavetable(wavetable_len, np.sin, cycle=2, phase=random.uniform(0, 1))
                #     wt.data = torch.cat([wt[:-1], wt[0].unsqueeze(-1)], dim=-1)
                #     wt.requires_grad = is_initial_wt_trainable
                # else:
                wt.data = torch.cat([wt[:-1], wt[0].unsqueeze(-1)], dim=-1)
                wt.requires_grad = is_initial_wt_trainable
        else:
            self.wavetables = wavetables
        
        self.attention = nn.Parameter(torch.ones(n_wavetables,).cuda())
        self.sr = sr
        self.block_size = block_size
        self.attention_softmax = nn.Softmax(dim=0)
        self.duration_secs = duration_secs

    def forward(self, pitch, amplitude):
        output_waveform_lst = []
        for wt_idx in range(len(self.wavetables)):
            wt = self.wavetables[wt_idx]
            if wt_idx not in [0, 1, 2, 3]:
                wt = nn.Tanh()(wt)  # ensure wavetable range is between [-1, 1]
            waveform = wavetable_osc(wt, pitch, self.sr)
            output_waveform_lst.append(waveform)

        # apply attention 
        attention = self.attention_softmax(self.attention)
        attention_output = attention
        attention = torch.stack(100 * self.duration_secs * [attention], dim=-1)
        attention_upsample = upsample(attention.unsqueeze(-1), self.block_size).squeeze()

        output_waveform = torch.stack(output_waveform_lst, dim=1)
        output_waveform = output_waveform * attention_upsample
        output_waveform_after = torch.sum(output_waveform, dim=1)
      
        output_waveform_after = output_waveform_after.unsqueeze(-1)
        output_waveform_after = output_waveform_after * amplitude
       
        return output_waveform_after, attention_output


class WavetableSynthV2(nn.Module):
    """
    take wavetable as input, not model parameters
    """
    def __init__(self,
                 sr=44100,
                 duration_secs=4,
                 block_size=160,
                 enable_amplitude=True):
        """
        Turn on smoothing to reduce noise in learnt wavetables.
        Smoothing takes in a 0-1 value, which is window size ratio w.r.t. wavetable length
        Also a max_smooth_window_size is specified
        """
        super(WavetableSynthV2, self).__init__()       
        self.sr = sr
        self.block_size = block_size
        self.duration_secs = duration_secs
        self.enable_amplitude = enable_amplitude

    def forward(self, pitch, amplitude, wavetables, attention):
        """
        batch size version
        input:
        wavetables: (bs, n_wavetables, wavetable_len), -1 to 1
        attention: softmax-ed, (bs, n_wavetables,)
        smoothing_coeff: (bs, ), 0 to 1

        output:
        (bs, dur * sr)
        """
        output_waveform_lst = []
        for wt_idx in range(wavetables.shape[1]):
            wt = wavetables[:, wt_idx, :]
            waveform = wavetable_osc_v2(wt, pitch, self.sr)

            output_waveform_lst.append(waveform)

        # apply attention 
        attention_upsample = torch.stack(100 * self.duration_secs * [attention], dim=-1)
        attention_upsample = upsample(torch.permute(attention_upsample, (1, 2, 0)), self.block_size)
        if (attention_upsample.shape[0] != 1):
            attention_upsample = attention_upsample.squeeze()   # TODO: a little hacky code here, need to remove
        attention_upsample = torch.permute(attention_upsample, (2, 0, 1))

        output_waveform = torch.stack(output_waveform_lst, dim=1)
        output_waveform = output_waveform * attention_upsample
        output_waveform_after = torch.sum(output_waveform, dim=1)
      
        output_waveform_after = output_waveform_after.unsqueeze(-1)
        if self.enable_amplitude:
            output_waveform_after = output_waveform_after * amplitude
       
        return output_waveform_after, attention


if __name__ == "__main__":
    # create a sine wavetable and to a simple synthesis test
    wavetable_len = 512
    sr = 16000
    duration = 4
    freq_t_1 = [739.99 for _ in range(sr)] + [523.25 for _ in range(sr)] + [349.23 for _ in range(sr * 2)]
    freq_t_1 = torch.tensor(freq_t_1)
    freq_t_2 = [523.25 for _ in range(sr)] + [349.23 for _ in range(sr)] + [739.99 for _ in range(sr * 2)]
    freq_t_2 = torch.tensor(freq_t_2)
    freq_t_3 = [349.23 for _ in range(sr)] + [739.99 for _ in range(sr)] + [523.25 for _ in range(sr * 2)]
    freq_t_3 = torch.tensor(freq_t_3)

    pitch, onset_frames, times = np.load("pitch.npy"), np.load("onset.npy"), np.load("times.npy")
    pitch = torch.tensor(pitch)
    pitch = upsample(pitch.unsqueeze(-1).unsqueeze(0), 160).squeeze()

    freq_t = torch.stack([pitch, pitch, pitch], dim=0)
    sine_wavetable = generate_wavetable(wavetable_len, np.sin)
    from utils import sawtooth_waveform
    saw_wavetable = generate_wavetable(wavetable_len, sawtooth_waveform)
    square_wavetable = generate_wavetable(wavetable_len, square_waveform)

    wavetable = torch.stack([sine_wavetable, saw_wavetable, square_wavetable], dim=0)

    # test batch wavetable_osc
    signal = wavetable_osc_v2(wavetable, freq_t, sr)

    # test with adsr
    shaper = ADSREnvelopeShaper()
    adsr = get_amp_shaper_v2(shaper, times[onset_frames], 
                            attack_secs=torch.tensor([0.00]),
                            decay_secs=torch.tensor([0.05]),
                            sustain_level=torch.tensor([0.0]))
    if adsr.shape[0] < 400:
        append_tensor = torch.tensor([adsr[-1]] * (400 - adsr.shape[0]))
        adsr = torch.cat([adsr, append_tensor], dim=-1)
    else:
        adsr = adsr[:400]
    adsr = upsample(adsr.unsqueeze(-1).unsqueeze(0), 160).squeeze()

    signal = signal * adsr
    
    # wt_synth = WavetableSynth(wavetables=sine_wavetable, sr=sr, duration_secs=4)
    # amplitude_t = torch.ones(sr * duration,)
    # amplitude_t = torch.stack([amplitude_t, amplitude_t, amplitude_t], dim=0)
    # amplitude_t = amplitude_t.unsqueeze(-1)

    # y = wt_synth(freq_t, amplitude_t, duration)
    # print(y.shape, 'y')
    # plt.plot(y.squeeze()[0].detach().numpy())
    # plt.show()
    sf.write('test_3s_v1.wav', signal.squeeze()[0].detach().numpy(), sr, 'PCM_24')
    sf.write('test_3s_v2.wav', signal.squeeze()[1].detach().numpy(), sr, 'PCM_24')
    sf.write('test_3s_v3.wav', signal.squeeze()[2].detach().numpy(), sr, 'PCM_24')


import os
from syntheon import infer_params


def test_dexed_inferencer():
    """
    just check if everything runs well for Dexed
    """
    output_params_file, eval_dict = infer_params(
        "test/test_audio/dexed_test_audio_1.wav", 
        "dexed", 
        enable_eval=True
    )
    assert os.path.exists(output_params_file)

    os.remove(output_params_file)


def test_vital_inferencer_1():
    """
    just check if everything runs well for Vital
    """
    loss_lst = [0.42, 0.11, 0.37, 0.06, 0.42, 0.18, 0.15]
    for idx in range(1, 8):
        output_params_file, eval_dict = infer_params(
            "test/test_audio/vital_test_audio_{}.wav".format(idx), 
            "vital", 
            enable_eval=True
        )
        assert os.path.exists(output_params_file)
        assert eval_dict["loss"] < loss_lst[idx - 1]
        os.remove(output_params_file)


from torch.utils.data import Dataset
import h5py
import numpy as np
import random
import torch
import math

class h5Dataset(Dataset):
    def __init__(self, sr, data_path,input_keys,max_audio_val=1,device='cpu'):
        self.sr = sr
        self.data_path = data_path
        self.input_data_dicts,self.dataset_len = self.cache_data(self.data_path,len(input_keys))
        self.max_audio_val = max_audio_val
        self.input_keys = input_keys
        self.device = device

    def cache_data(self, data_path,nfeatures):
        '''
        Load data to dictionary in RAM
        '''
        h5f = h5py.File(data_path, 'r')
        cache = {}
        keys = h5f.keys()
        nkeys = len(keys)
        ndata = (len(keys)//nfeatures)
        if((nkeys//nfeatures)*nfeatures != nkeys):
            raise Exception("Unexpected dataset len.")

        for key in keys:
            cache[key] = np.array(h5f[key])
        h5f.close()

        return cache, ndata

    def __getitem__(self, idx):
        #print("[DEBUG] __getitem__ fetching: {}".format(idx))

        #Generate current item keys to fetch from RAM cache
        item_keys = [f'{idx}_{k}' for k in self.input_keys ]

        # Load dictionary
        x = {}
        for v,k in enumerate(self.input_keys):
            x[k] = torch.tensor(self.input_data_dicts[item_keys[v]]).unsqueeze(-1).to(self.device)

        #for k in x.keys():
        #    print(f'{k}: {x[k].shape} ',end='')
        #print('')

        return x

    def __len__(self):
        return self.dataset_len



from syntheon.inferencer.inferencer import Inferencer, InferenceInput, InferenceOutput
from syntheon.inferencer.vital.models.model_v2 import WTSv2
from syntheon.inferencer.vital.models.preprocessor import *
from syntheon.inferencer.vital.models.core import multiscale_fft
from syntheon.converter.vital.vital_constants import N_WAVETABLES, CUSTOM_KEYS
import yaml 
import torch
import numpy as np
import json

with open("syntheon/inferencer/vital/config.yaml", 'r') as stream:
    config = yaml.safe_load(stream)

# general parameters
sr = config["common"]["sampling_rate"]
block_size = config["common"]["block_size"]
duration_secs = config["common"]["duration_secs"]
batch_size = config["train"]["batch_size"]
scales = config["train"]["scales"]
overlap = config["train"]["overlap"]
hidden_size = config["train"]["hidden_size"]
n_harmonic = config["train"]["n_harmonic"]
n_bands = config["train"]["n_bands"]
n_wavetables = config["train"]["n_wavetables"]
n_mfcc = config["train"]["n_mfcc"]
train_lr = config["train"]["start_lr"]
visualize = config["visualize"]
device = config["device"]
signal_length = sr * 4


class VitalInferenceOutput(InferenceOutput):
    def __init__(self):
        InferenceOutput.__init__(self)
        self.wt_output = None           # TODO: can put default values here
        self.attention_output = None
        self.attack = None
        self.decay = None
        self.sustain = None


class VitalInferenceInput(InferenceInput):
    def __init__(self):
        self.y = None
        self.pitch = None
        self.loudness = None
        self.times = None
        self.onset_frames = None
        self.mfcc = None


class VitalInferencer(Inferencer):
    def convert(self, audio_fname, model_pt_fname=None, enable_eval=False):
        # for vital, the model loading depends on signal input length
        if model_pt_fname is None:
            model_pt_fname = "syntheon/inferencer/vital/checkpoints/model_adsr_loudness_v2.pt"
        
        y, pitch, loudness, times, onset_frames, mfcc = preprocess(audio_fname, sampling_rate=16000, block_size=160, 
                                                                   signal_length=signal_length)
        inference_input = VitalInferenceInput()
        inference_input.y = y
        inference_input.pitch = pitch
        inference_input.loudness = loudness
        inference_input.times = times
        inference_input.onset_frames = onset_frames
        inference_input.mfcc = mfcc

        model = self.load_model(model_pt_fname, self.device)
        inference_output = self.inference(model, inference_input, self.device, enable_eval=enable_eval)
        synth_params_dict = self.convert_to_preset(inference_output)
        return synth_params_dict, inference_output.eval_dict

    def load_model(self, model_pt_fname, device="cuda"):
        model = WTSv2(hidden_size=hidden_size, n_harmonic=n_harmonic, n_bands=n_bands, sampling_rate=sr,
                block_size=block_size,  mode="wavetable", 
                duration_secs=4, num_wavetables=1, wavetable_smoothing=False, preload_wt=True, enable_amplitude=False,
                is_round_secs=False, device=device)
        if device == "cuda":
            model.load_state_dict(torch.load(model_pt_fname))
            model.cuda()
        else:
            model.load_state_dict(torch.load(model_pt_fname, map_location=torch.device('cpu')))
        model.eval()        
        return model
    
    def inference(self, model, inference_input, device="cuda", enable_eval=False):
        if device == "cuda":
            inference_input.y = inference_input.y.cuda()
            inference_input.mfcc = inference_input.mfcc.cuda()
            inference_input.pitch = inference_input.pitch.cuda()
            inference_input.loudness = inference_input.loudness.cuda()

        # forward pass
        with torch.no_grad():
            _, adsr, output, attention_output, wavetables, _, _ = model(
                inference_input.y, 
                inference_input.mfcc, 
                inference_input.pitch, 
                inference_input.loudness, 
                inference_input.times, 
                inference_input.onset_frames
            )

        # write wavetables to numpy file
        wt_output = []

        # interp from 512 to 2048
        output_length = 2048
        for i in range(N_WAVETABLES):
            wt = wavetables[i].cpu().detach().numpy().squeeze()
            wt_interp = np.interp(
                np.linspace(0, 1, output_length, endpoint=False),
                np.linspace(0, 1, wt.shape[0], endpoint=False),
                wt,
            )
            wt_output.append(wt_interp)

        wt_output = np.stack(wt_output, axis=0)
        attention_output = attention_output.cpu().detach().numpy().squeeze()

        inference_output = VitalInferenceOutput()
        inference_output.wt_output = wt_output
        inference_output.attention_output = attention_output
        inference_output.attack = adsr[0][0].cpu().detach().numpy().squeeze().item()
        inference_output.decay = adsr[1][0].cpu().detach().numpy().squeeze().item()
        inference_output.sustain = adsr[2][0].cpu().detach().numpy().squeeze().item()

        if enable_eval:
            self.eval(inference_input.y, output, inference_output)

        return inference_output
    
    def convert_to_preset(self, inference_output):
        with open("syntheon/inferencer/vital/init.vital") as f:
            x = json.load(f)

        x[CUSTOM_KEYS] = {}
        x[CUSTOM_KEYS]["wavetables"] = []
        for idx in range(N_WAVETABLES):
            cur_dict = {
                "name": "Litmus WT {}".format(idx + 1),
                "wavetable": inference_output.wt_output[idx],
                "osc_level": inference_output.attention_output[idx].item()
            }
            x[CUSTOM_KEYS]["wavetables"].append(cur_dict)
            x[CUSTOM_KEYS]["adsr"] = {}
            x[CUSTOM_KEYS]["adsr"]["attack"] = inference_output.attack
            x[CUSTOM_KEYS]["adsr"]["attack_power"] = 0.0
            x[CUSTOM_KEYS]["adsr"]["decay"] = inference_output.decay
            x[CUSTOM_KEYS]["adsr"]["decay_power"] = 0.0
            x[CUSTOM_KEYS]["adsr"]["sustain"] = inference_output.sustain

        return x
    
    def eval(self, y, output, inference_output):
        ori_stft = multiscale_fft(
                    y[0].squeeze(),
                    scales,
                    overlap,
                )
        rec_stft = multiscale_fft(
            output[0].squeeze(),
            scales,
            overlap,
        )

        loss = 0
        for s_x, s_y in zip(ori_stft, rec_stft): 
            lin_loss = ((s_x - s_y).abs()).mean()
            loss += lin_loss
        
        inference_output.eval_dict["loss"] = loss.item()
        inference_output.eval_dict["output"] = output[0].cpu().detach().numpy().squeeze()


if __name__ == "__main__":
    # TODO: move to test folder
    vital_inferencer = VitalInferencer(device="cpu")
    params, eval_dict = vital_inferencer.convert("test/test_audio/vital_test_audio_2.wav", enable_eval=True)

    from syntheon.converter.vital.vital_converter import VitalConverter
    vital_converter = VitalConverter()
    vital_converter.dict = params
    vital_converter.parseToPluginFile("vital_output.vital")

    


"""
For loading and preprocessing audio
"""
import numpy as np
import torch
from syntheon.inferencer.vital.models.core import extract_loudness, extract_pitch
import librosa
import yaml 
from nnAudio import Spectrogram

with open("syntheon/inferencer/vital/config.yaml", 'r') as stream:
    config = yaml.safe_load(stream)

# general parameters
sr = config["common"]["sampling_rate"]
n_mfcc = config["train"]["n_mfcc"]

spec = Spectrogram.MFCC(sr=sr, n_mfcc=n_mfcc)


def sanitize_onsets(times, onset_frames, onset_strengths):
    """
    times: actual timestamp per frame in STFT (by hop length)
            e.g. [0, 0.3, 0.6, ...]
    onset_frames: index list, index on `times` to get an onset event
    onset_strengths: get strength per frame. same shape as times.
                        so when strength is high will have a onset event, index in `onset_frames`  
    """
    # TODO: need to check if we need this always
    res_frames = []

    cur_frame = onset_frames[0]
    cur_time = times[cur_frame]
    res_frames.append(cur_frame)

    for frame in onset_frames[1:]:
        if times[frame] - cur_time > 0.05:  # TODO: parameterize
            res_frames.append(frame)
        cur_frame = frame
        cur_time = times[frame]

    return np.array(res_frames)


def aggregate(vals):
    """
    aggregate the window of pitch values.
    rationale: bin pitch values (to reduce fluctuation), get the bin with most values within the window
    """
    bins = {}
    for val in vals:
        bin = val // 10
        if bin in bins:
            bins[bin].append(val)
        else:
            bins[bin] = [val]
    
    sorted_bins = sorted(bins.keys())
    max_len_bin = sorted_bins[0]

    for bin in sorted_bins:
        if len(bins[bin]) > len(bins[max_len_bin]):
            max_len_bin = bin
    
    return bins[max_len_bin][0]



def monotonize_pitch(times, onset_frames, pitch):
    """
    remove wobbling frequencies in pitch. take the pitch value on the onset frame
    problem is accuracy issue -- need to align onset and pitch
    because librosa onset might read wrong pitch from crepe output
    """
    res_pitch = np.zeros(pitch.shape)
    pitch_map_lst = []

    prev_ts = times[onset_frames[0]]

    for idx, frame in enumerate(onset_frames):
        if idx == 0:
            continue
        ts = times[frame]
        pitch_vals = pitch[int(prev_ts * 100) : int(ts * 100)]

        if len(pitch_vals) > 0:
            cur_pitch = aggregate(pitch_vals)
            pitch_map_lst.append((int(prev_ts * 100), cur_pitch))
            prev_ts = ts

    # for final frame
    ts = times[-1]
    pitch_vals = pitch[int(prev_ts * 100) : int(ts * 100)]
    if len(pitch_vals) > 0:
        cur_pitch = aggregate(pitch_vals)
        pitch_map_lst.append((int(prev_ts * 100), cur_pitch))
    
    if pitch_map_lst[0][0] == 0:
        res_pitch[0] = pitch_map_lst[0][1]
        cur_pitch = pitch_map_lst[0][1]
        cur_idx = 1
    else:
        res_pitch[0] = 0
        cur_pitch = 0
        cur_idx = 0
    
    for i in range(1, len(pitch)):
        if i == pitch_map_lst[cur_idx][0]:
            cur_pitch = pitch_map_lst[cur_idx][1]
            res_pitch[i] = cur_pitch
            if cur_idx < len(pitch_map_lst) - 1:
                cur_idx += 1
        else:
            res_pitch[i] = cur_pitch

    return res_pitch
    

def preprocess(f, sampling_rate, block_size, signal_length=-1, oneshot=True):
    x, sr = librosa.load(f, sampling_rate)
    if signal_length == -1:     # full length
        signal_length = len(x)
    else:
        if len(x) > signal_length:
            x = x[:signal_length*sampling_rate]
        elif len(x) < signal_length:
            N = (signal_length - len(x) % signal_length) % signal_length
            x = np.pad(x, (0, N))

        if oneshot:
            x = x[..., :signal_length]

    D = np.abs(librosa.stft(x))
    times = librosa.times_like(D, sr=sr)
    onset_strengths = librosa.onset.onset_strength(y=x, sr=sr, aggregate=np.median)
    onset_frames = librosa.onset.onset_detect(y=x, sr=sr)

    onset_frames = sanitize_onsets(times, onset_frames, onset_strengths)

    # TODO: HACK for now, onset detector missed. not all samples need this!!
    onset_frames = np.concatenate([np.array([0]), onset_frames])

    pitch = extract_pitch(x, sampling_rate, block_size)
    loudness = extract_loudness(x, sampling_rate, block_size)

    pitch_monotonize = monotonize_pitch(times, onset_frames, pitch)
    pitch = pitch_monotonize
    x = x.reshape(-1, signal_length)
    pitch = pitch.reshape(x.shape[0], -1).squeeze()
    loudness = loudness.reshape(x.shape[0], -1)

    # prepare for inference input
    x = torch.tensor(x)
    pitch = torch.tensor(pitch).unsqueeze(0)
    loudness = torch.tensor(loudness)

    x = torch.cat([x, x], dim=0)
    pitch = torch.cat([pitch, pitch], dim=0)
    loudness = torch.cat([loudness, loudness], dim=0)

    mean_loudness, std_loudness = -39.74668743704927, 54.19612404969509
    pitch, loudness = pitch.unsqueeze(-1).float(), loudness.unsqueeze(-1).float()
    loudness = (loudness - mean_loudness) / std_loudness

    mfcc = spec(x)

    return x, pitch, loudness, times, onset_frames, mfcc



import ddx7.core as core
import torch
import torch.nn as nn
import torchcrepe
from functools import partial


'''
Asimetric L1 distance
'''
def asim_l1_distance(a,b,alpha=1,beta=1):
    diff = a-b
    pos_diff = diff * (diff > 0)
    neg_diff = diff * (diff < 0)
    as_diff = alpha * pos_diff + beta * neg_diff
    as_mse = torch.abs(as_diff).mean()
    return as_mse


def asim_msfft_loss(a1,
                    a2,
                    scales=[4096, 2048, 1024, 512, 256, 128],
                    overlap=0.75,
                    alpha=1,
                    beta=1):
    '''
    DDSP Original MS FFT loss with lin + log spectra analysis
    '''
    if(len(a1.size()) == 3):
        a1 = a1.squeeze(-1)
    if(len(a2.size()) == 3):
        a2 = a2.squeeze(-1)
    ori_stft = core.multiscale_fft(
        a1,
        scales,
        overlap,
    )
    rec_stft = core.multiscale_fft(
        a2,
        scales,
        overlap,
    )

    loss = 0
    for s_x, s_y in zip(ori_stft, rec_stft):
        lin_loss = asim_l1_distance(s_x, s_y,alpha,beta)
        log_loss = asim_l1_distance(core.safe_log(s_x),core.safe_log(s_y),alpha,beta)
        loss = loss + lin_loss + log_loss

    return loss



def ddsp_msfft_loss(a1,
                    a2,
                    scales=[4096, 2048, 1024, 512, 256, 128],
                    overlap=0.75):
    '''
    DDSP Original MS FFT loss with lin + log spectra analysis
        Some remarks: the stfts have to be normalized otherwise the netowrk weights different excerpts to different importance.
                      We compute the mean of the L1 difference between normalized magnitude spectrograms
                      so that the magnitude of the loss do not change with the window size.
    '''
    if(len(a1.size()) == 3):
        a1 = a1.squeeze(-1)
    if(len(a2.size()) == 3):
        a2 = a2.squeeze(-1)
    ori_stft = core.multiscale_fft(
        a1,
        scales,
        overlap,
    )
    rec_stft = core.multiscale_fft(
        a2,
        scales,
        overlap,
    )

    loss = 0
    for s_x, s_y in zip(ori_stft, rec_stft):
        lin_loss = (s_x - s_y).abs().mean()
        log_loss = (core.safe_log(s_x) - core.safe_log(s_y)).abs().mean()
        loss = loss + lin_loss + log_loss

    return loss

class rec_loss(nn.Module):
    def __init__(self,scales,overlap,alpha=None,beta=None):
        super().__init__()
        self.scales = scales
        self.overlap = overlap
        if(alpha is not None and beta is not None):
            self.loss_fn = partial(asim_msfft_loss,alpha=alpha,beta=beta)
            print(f'[INFO] rec_loss() - Using asimetrical reconstruction loss. alpha: {alpha} - beta: {beta}')
        else:
            self.loss_fn = ddsp_msfft_loss
    def forward(self,ref,synth):
        return self.loss_fn(ref,synth,
                    self.scales,
                    self.overlap)



"""
ADSR envelope shaper
"""
import numpy as np
import torch
import matplotlib.pyplot as plt
from torch import nn
import yaml


with open("syntheon/inferencer/vital/config.yaml", 'r') as stream:
    config = yaml.safe_load(stream)
device = config["device"]


def soft_clamp_min(x, min_v, T=100):
    return torch.sigmoid((min_v-x)*T)*(min_v-x)+x


class DiffRoundFunc(torch.autograd.Function):
    @staticmethod
    def forward(ctx, input):
        ctx.input = input
        return torch.round(input * 10 ** 2) / (10 ** 2)     # because 2 decimal point, 0.01 is the minimum ratio

    @staticmethod
    def backward(ctx, grad_output):
        grad_input = grad_output.clone()
        return grad_input


class ADSREnvelopeShaper(nn.Module):
    def __init__(self, is_round_secs=False):
        super(ADSREnvelopeShaper, self).__init__()
        self.attack_secs = torch.tensor([0])
        self.attack_power = torch.tensor([0])
        self.decay_secs = torch.tensor([0])
        self.decay_power = torch.tensor([0])
        self.sustain_level = torch.tensor([0])
        self.release_secs = torch.tensor([0])
        self.release_power = torch.tensor([0])

        self.is_round_secs = is_round_secs
        self.round_decimal_places = 2   # because block size = 100, so min resolution = 1/ 100
    
    def power_function(self, x, pow=2):
        if pow > 0: # convex
            # transpose
            plt.plot(x.squeeze().detach().numpy(), label='test1')

            if x.squeeze()[0] > x.squeeze()[-1]:
                y_intercept = x.squeeze()[-1]
                y = x - x[:, -1, :]
                max_val = y.squeeze()[0]
                y = y / max_val
            else:
                y_intercept = x.squeeze()[0]
                y = x - x[:, 0, :]
                max_val = y.squeeze()[-1]
                y = y / max_val
            
            y = y ** pow

            # transpose back
            y = y * max_val + y_intercept

            plt.plot(y.squeeze().detach().numpy(), label='test3')
            plt.show()

        else:
            # transpose
            plt.plot(x.squeeze().detach().numpy(), label='test1')

            if x.squeeze()[0] > x.squeeze()[-1]:
                max_val = x.squeeze()[0]
                y = x - x[:, 0, :]
                y_intercept = y.squeeze()[-1]
                y = y / -y_intercept
            else:
                max_val = x.squeeze()[-1]
                y = x - x[:, -1, :]
                y_intercept = y.squeeze()[0]
                y = y / -y_intercept

            plt.plot(y.squeeze().detach().numpy(), label='test2')

            y = -(y ** -pow)

            plt.plot(y.squeeze().detach().numpy(), label='test3')

            # transpose back
            y = y * -y_intercept + max_val

            plt.plot(y.squeeze().detach().numpy(), label='test4')
            plt.legend()
            plt.show()

            # plt.plot(y.squeeze().detach().numpy(), label='test3')
            # plt.show()

        return y
    
    def gen_envelope(self, attack, decay, sus_level, release,
                     floor=None, peak=None, n_frames=250, pow=2):
        """generate envelopes from parameters
        Args:
            floor (torch.Tensor): floor level of the signal 0~1, 0=min_value (batch, 1, channels)
            peak (torch.Tensor): peak level of the signal 0~1, 1=max_value (batch, 1, channels)
            attack (torch.Tensor): relative attack point 0~1 (batch, 1, channels)
            decay (torch.Tensor): actual decay point is attack+decay (batch, 1, channels)
            sus_level (torch.Tensor): sustain level 0~1 (batch, 1, channels)
            release (torch.Tensor): release point is attack+decay+release (batch, 1, channels)
            note_off (float or torch.Tensor, optional): note off position. Defaults to 0.8.
            n_frames (int, optional): number of frames. Defaults to None.
        Returns:
            torch.Tensor: envelope signal (batch_size, n_frames, 1)
        """
        if floor is None:
            floor = torch.tensor([0.]).unsqueeze(0).unsqueeze(-1)
            if device == "cuda":
                floor = floor.cuda()
        if peak is None:
            peak = torch.tensor([1.]).unsqueeze(0).unsqueeze(-1)
            if device == "cuda":
                peak = peak.cuda()
        
        # attack ratio can be 0, but the initial adsr 0 values should be epsilon-ed
        # decay ratio should be larger than 1 / total_n_frames
        # decay secs should be larger than 1 / block_size
        # attack and decay secs will be rounded up to minimum resolution (min resolution = 1/ block_size)
        # DO WE REALLY NEED THIS? 
        MIN_RATIO = 1 / 400     # HACK, 4 seconds * block size 100
        
        attack = torch.clamp(attack, min=0, max=1)
        decay = torch.clamp(decay, min=0, max=1)
        sus_level = torch.clamp(sus_level, min=0.001, max=1)
        release = torch.clamp(release, min=0, max=1)

        batch_size = attack.shape[0]
        if n_frames is None:
            n_frames = self.n_frames
        # batch, n_frames, 1
        x = torch.linspace(0, 1.0, n_frames)[None, :, None].repeat(batch_size, 1, 1)
        x[:, 0, :] = 1e-6       # offset 0 to epsilon value, so when attack = 0, first adsr value is not 0 but 1
        x = x.to(attack.device)

        A = x / (attack + 1e-6)
        # A = self.power_function(A, pow=2)
        A = torch.clamp(A, max=1.0)

        D = (x - attack) * (sus_level - 1) / (decay+1e-6)
        # D = self.power_function(D, pow=-2.7)
        D = torch.clamp(D, max=0.0)
        D = soft_clamp_min(D, sus_level-1)

        S = (x - 1) * (-sus_level / (release+1e-6))
        S = torch.clamp(S, max=0.0)
        S = soft_clamp_min(S, -sus_level)

        signal = (A + D + S) * (peak - floor) + floor
        return torch.clamp(signal, min=0., max=1.)
    
    def forward(self, 
                attack_secs,
                decay_secs,
                sustain_level,
                block_size=100, 
                sr=44100, 
                total_secs=8):
        if self.is_round_secs:
            attack_secs = DiffRoundFunc.apply(attack_secs)
            decay_secs = DiffRoundFunc.apply(decay_secs)

        self.attack_secs = attack_secs
        self.decay_secs = decay_secs
        self.sustain_level = sustain_level
        
        attack_ratio = attack_secs / total_secs
        decay_ratio = decay_secs / total_secs
        # TODO: parameteize release_ratio
        release_ratio = torch.tensor([0.]).repeat(attack_secs.size(0), 1, 1)
        if device == "cuda":
            release_ratio = release_ratio.cuda()

        attack_ratio = attack_ratio.unsqueeze(-1).unsqueeze(-1)
        decay_ratio = decay_ratio.unsqueeze(-1).unsqueeze(-1)
        sus_level = sustain_level.unsqueeze(-1).unsqueeze(-1)

        signal = self.gen_envelope(attack_ratio, decay_ratio, sus_level, release_ratio,
                                    floor=None, peak=None, n_frames=int(total_secs * block_size),
                                    pow=2)
        return signal.squeeze()


def get_amp_shaper(
                shaper, 
                onsets, 
                attack_secs,
                decay_secs,
                sustain_level,
                offsets=None):
    """
    implement case with no offset first.
    """
    if offsets is None:
        # if offset not specified, take next onset as offset
        offsets = onsets[1:]
        onsets = onsets[:len(onsets) - 1]

    start_offset = int(onsets[0] * 100)        # TODO: 100 is block size
    onsets, offsets = torch.tensor(onsets), torch.tensor(offsets)
    if device == "cuda":
        onsets, offsets = onsets.cuda(), offsets.cuda()
    dur_vec = offsets - onsets
    lst = []

    # append zeros first before first onset
    if device == "cuda":
        lst.append(torch.zeros(start_offset).cuda())
    else:
        lst.append(torch.zeros(start_offset))

    for dur in dur_vec:
        dur = round(dur.item(), 2)
        adsr = shaper(
            attack_secs=torch.tensor([0.2, 0.1]), 
            decay_secs=torch.tensor([0.1, 0.2]),
            sustain_level=torch.tensor([0.9, 0.2]),
            total_secs=dur)
        lst.append(adsr[0].squeeze())   # TODO: fix the batch size case
    
    final_signal = torch.cat(lst, dim=0)
    return final_signal


def get_amp_shaper_v2(
                shaper, 
                onsets, 
                attack_secs,
                decay_secs,
                sustain_level,
                offsets=None):
    """
    implement case with no offset first. enable batches
    """
    if offsets is None:
        # if offset not specified, take next onset as offset
        offsets = onsets[1:]
        onsets = onsets[:len(onsets) - 1]

    start_offset = int(onsets[0] * 100)        # TODO: 100 is block size
    onsets, offsets = torch.tensor(onsets), torch.tensor(offsets)
    if device == "cuda":
        onsets, offsets = onsets.cuda(), offsets.cuda()
    dur_vec = offsets - onsets
    lst = []

    # append zeros first before first onset
    if device == "cuda":
        lst.append(torch.zeros(start_offset).cuda())
    else:
        lst.append(torch.zeros(start_offset))

    for dur in dur_vec:
        dur = round(dur.item(), 2)
        adsr = shaper(
            attack_secs=attack_secs, 
            decay_secs=decay_secs,
            sustain_level=sustain_level,
            total_secs=dur)
        
        # adsr shape should be (bs, dur * block_size)
        lst.append(adsr)
    
    final_signal = torch.cat(lst, dim=-1)
    return final_signal


if __name__ == "__main__":
    # TODO: unit test for this class
    shaper = ADSREnvelopeShaper(is_round_secs=False)
    adsrs = []
    for elem in [0.0, 0.001, 0.005, 0.01, 0.02]:
        attack_secs, decay_secs, sustain_level = torch.tensor([0.2]), torch.tensor([elem]), torch.tensor([0.8])
        if device == "cuda": 
            attack_secs = attack_secs.cuda()
            decay_secs = decay_secs.cuda()
            sustain_level = sustain_level.cuda()
        
        x2 = shaper(
            attack_secs=attack_secs, 
            decay_secs=decay_secs,
            sustain_level=sustain_level,
            total_secs=4)

        adsrs.append(x2.squeeze().cpu().detach().numpy()[:30])

    for idx, elem in enumerate([0.0, 0.001, 0.005, 0.01, 0.02]):
        plt.plot(adsrs[idx], label=str(elem))
        plt.scatter(range(30), adsrs[idx])
    plt.legend()
    plt.show()


"""
DX7 syx -> json constants.

Credits to source: https://github.com/Nintorac/NeuralDX7/blob/master/scratch/dx7_constants.py
DX7 specification: https://github.com/asb2m10/dexed/blob/master/Documentation/sysex-format.txt
"""
import bitstruct

N_OSC = 6
N_VOICES = 32

GLOBAL_VALID_RANGES = {
    'PR1':  range(0, 99+1),
    'PR2':  range(0, 99+1),
    'PR3':  range(0, 99+1),
    'PR4':  range(0, 99+1),
    'PL1':  range(0, 99+1),
    'PL2':  range(0, 99+1),
    'PL3':  range(0, 99+1),
    'PL4':  range(0, 99+1),
    'ALG':  range(0, 31+1),
    'OKS':  range(0, 1+1),
    'FB':   range(0, 7+1),
    'LFS':  range(0, 99+1),
    'LFD':  range(0, 99+1),
    'LPMD':  range(0, 99+1),
    'LAMD':  range(0, 99+1),
    'LPMS': range(0, 7+1),
    'LFW':  range(0, 5+1),
    'LKS':  range(0, 1+1),
    'TRNSP':  range(0, 48+1),
    'NAME CHAR 1': range(128),
    'NAME CHAR 2': range(128),
    'NAME CHAR 3': range(128),
    'NAME CHAR 4': range(128),
    'NAME CHAR 5': range(128),
    'NAME CHAR 6': range(128),
    'NAME CHAR 7': range(128),
    'NAME CHAR 8': range(128),
    'NAME CHAR 9': range(128),
    'NAME CHAR 10': range(128),
 }

OSCILLATOR_VALID_RANGES = {
    'R1':  range(0, 99+1),
    'R2':  range(0, 99+1),
    'R3':  range(0, 99+1),
    'R4':  range(0, 99+1),
    'L1':  range(0, 99+1),
    'L2':  range(0, 99+1),
    'L3':  range(0, 99+1),
    'L4':  range(0, 99+1),
    'BP':  range(0, 99+1),
    'LD':  range(0, 99+1),
    'RD':  range(0, 99+1),
    'RC':  range(0, 3+1),
    'LC':  range(0, 3+1),
    'DET': range(0, 14+1),
    'RS':  range(0, 7+1),
    'KVS': range(0, 7+1),
    'AMS': range(0, 3+1),
    'OL':  range(0, 99+1),
    'FC':  range(0, 31+1),
    'M':   range(0, 1+1),
    'FF':  range(0, 99+1),
}

VOICE_PARAMETER_RANGES = {f'{i}_{key}': value for key, value in OSCILLATOR_VALID_RANGES.items() for i in range(N_OSC)}
VOICE_PARAMETER_RANGES.update(GLOBAL_VALID_RANGES)


HEADER_KEYS = [
    'ID',
    'Sub-status',
    'format number',
    'byte count',
    'byte count',
]

GENERAL_KEYS = [
    'PR1',
    'PR2',
    'PR3',
    'PR4',
    'PL1',
    'PL2',
    'PL3',
    'PL4',
    'ALG',
    'OKS',
    'FB',
    'LFS',
    'LFD',
    'LPMD',
    'LAMD',
    'LPMS',
    'LFW',
    'LKS',
    'TRNSP',
    'NAME CHAR 1',
    'NAME CHAR 2',
    'NAME CHAR 3',
    'NAME CHAR 4',
    'NAME CHAR 5',
    'NAME CHAR 6',
    'NAME CHAR 7',
    'NAME CHAR 8',
    'NAME CHAR 9',
    'NAME CHAR 10',
]

OSC_KEYS = [
    'R1',
    'R2',
    'R3',
    'R4',
    'L1',
    'L2',
    'L3',
    'L4',
    'BP',
    'LD',
    'RD',
    'RC',
    'LC',
    'DET',
    'RS',
    'KVS',
    'AMS',
    'OL',
    'FC',
    'M',
    'FF',
]

FOOTER_KEYS = ['checksum']


VOICE_KEYS = [f'{i}_{key}' for i in range(6) for key in OSC_KEYS] + \
        GENERAL_KEYS 

KEYS =  HEADER_KEYS + \
        list(VOICE_KEYS * N_VOICES) + \
        FOOTER_KEYS

header_bytes = [
    'p1u7',             # ID # (i=67; Yamaha)
    'p1u7',             # Sub-status (s=0) & channel number (n=0; ch 1)
    'p1u7',             # format number (f=9; 32 voices)
    'p1u7',             # byte count MS byte
    'p1u7',             # byte count LS byte (b=4096; 32 voices)
]

general_parameter_bytes = [ 
    'p1u7',             # PR1
    'p1u7',             # PR2
    'p1u7',             # PR3
    'p1u7',             # PR4
    'p1u7',             # PL1
    'p1u7',             # PL2
    'p1u7',             # PL3
    'p1u7',             # PL4
    'p3u5',             # ALG
    'p4u1u3',           # OKS|    FB
    'p1u7',             # LFS
    'p1u7',             # LFD
    'p1u7',             # LPMD
    'p1u7',             # LAMD
    'p1u3u3u1',         # LPMS |      LFW      |LKS
    'p1u7',             # TRNSP
    'p1u7',             # NAME CHAR 1
    'p1u7',             # NAME CHAR 2
    'p1u7',             # NAME CHAR 3
    'p1u7',             # NAME CHAR 4
    'p1u7',             # NAME CHAR 5
    'p1u7',             # NAME CHAR 6
    'p1u7',             # NAME CHAR 7
    'p1u7',             # NAME CHAR 8
    'p1u7',             # NAME CHAR 9
    'p1u7',             # NAME CHAR 10
]

osc_parameter_bytes = [
    'p1u7',         # R1
    'p1u7',         # R2
    'p1u7',         # R3
    'p1u7',         # R4
    'p1u7',         # L1
    'p1u7',         # L2
    'p1u7',         # L3
    'p1u7',         # L4
    'p1u7',         # BP
    'p1u7',         # LD
    'p1u7',         # RD
    'p4u2u2',       # RC | LC 
    'p1u4u3',       # DET | RS
    'p3u3u2',       # KVS | AMS
    'p1u7',         # OL
    'p2u5u1',       # FC | M
    'p1u7'          # FF
]

voice_bytes = (osc_parameter_bytes * N_OSC) + general_parameter_bytes

tail_bytes = [
    'p1u7',         # checksum
]

full_string = ''.join(header_bytes + osc_parameter_bytes * 6 + general_parameter_bytes)
dx7_struct = bitstruct.compile(full_string)

voice_struct = bitstruct.compile(''.join(voice_bytes), names=VOICE_KEYS)
header_struct = bitstruct.compile(''.join(header_bytes))